import os
import tempfile
from typing import Optional

import pymongo_inmemory
from pymongo_inmemory.context import Context


class MightstoneInMemoryContext(Context):
    def __init__(
        self,
        data_dir: Optional[str] = None,
        cache_dir: Optional[str] = None,
        database: Optional[str] = None,
    ) -> None:
        super().__init__()

        if data_dir:
            os.makedirs(os.path.join(data_dir, "mongo", "data"), exist_ok=True)

        if database:
            self.dbname = database

        self.mongo_version = "7.0"
        self.mongod_data_folder = data_dir
        self.mongo_client_host = None
        self.storage_engine = "wiredTiger"

        if cache_dir:
            os.makedirs(os.path.join(cache_dir, "mongo", "download"), exist_ok=True)
            os.makedirs(os.path.join(cache_dir, "mongo", "extract"), exist_ok=True)
            os.makedirs(
                os.path.join(cache_dir, "mongo", "extract", self.url_hash),
                exist_ok=True,
            )
            os.makedirs(
                os.path.join(cache_dir, "mongo", "download", self.url_hash),
                exist_ok=True,
            )

            self.download_folder = os.path.join(cache_dir, "mongo", "extract")
            self.extract_folder = os.path.join(cache_dir, "mongo", "download")
            self.archive_folder = os.path.join(
                cache_dir, "mongo", "download", self.url_hash
            )
            self.extracted_folder = os.path.join(
                cache_dir, "mongo", "extract", self.url_hash
            )


class Mongod(pymongo_inmemory.Mongod):
    def __init__(self, pim_context: MightstoneInMemoryContext):
        super().__init__(pim_context)
        self.started = False

    def start(self):
        if self.started:
            return
        if self.is_locked:
            raise RuntimeError(
                "Mongo seems to be already running under another Mightstone instance. "
                "If you plan to use the storage concurrently, you should consider "
                "using a dedicated mongodb instance."
            )
        super().start()
        self.started = True

    def cleanup(self):
        super().stop()

    @property
    def is_locked(self):
        lock_file = os.path.join(self.data_folder, "mongod.lock")
        if not os.path.exists(lock_file):
            return False

        return os.path.getsize(lock_file) != 0



import datetime
import importlib
import inspect
import pkgutil
import sys
from abc import ABC
from types import ModuleType
from typing import Any, Optional, Type, Union
from uuid import UUID

from beanie import Document, PydanticObjectId
from beanie.exceptions import CollectionWasNotInitialized
from beanie.odm.settings.document import DocumentSettings
from pydantic import (
    BaseModel,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)
from pydantic_extra_types.color import Color


def Fallback(fallback_value: Any):
    def use_fallback(
        v: Any,
        handler: ValidatorFunctionWrapHandler,
        info: ValidationInfo,
    ) -> Any:
        try:
            return handler(v)
        except ValueError:
            return fallback_value

    return WrapValidator(use_fallback)


class MightstoneModel(ABC, BaseModel):
    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))


class MightstoneDocument(MightstoneModel):
    def to_serializable(
        self,
    ):
        target_class = get_serializable_class(self.__class__)
        if not target_class:
            raise Exception(
                "There is no available serializable variant of %s"
                % self.__class__.__name__
            )

        return target_class.model_validate(self.model_dump())


class MightstoneSerializableDocument(MightstoneModel, Document):
    id: Optional[Union[UUID, PydanticObjectId]] = None  # type: ignore

    @classmethod
    def get_settings(cls) -> DocumentSettings:
        try:
            return super().get_settings()
        except CollectionWasNotInitialized as e:
            raise MightstoneError(
                "In order to benefit from database serialization you need to "
                "initialize Mightstone through Mighstone.with_beanie() factory"
                "or use Mighstone.beanie_init()."
            ) from e


def get_serializable_class(
    cls: type[MightstoneDocument],
) -> Union[MightstoneSerializableDocument, None]:
    for name, candidate in inspect.getmembers(
        sys.modules[cls.__module__], inspect.isclass
    ):
        if candidate in [
            Document,
            MightstoneDocument,
            MightstoneSerializableDocument,
            cls,
        ]:
            continue
        if issubclass(candidate, (cls, Document)):
            return candidate  # type: ignore
    return None


def get_documents(root_module: Union[ModuleType, str] = "mightstone"):
    """
    Explore a package (mightstone by default) to find any file that extends Document

    This is useful to initialize beanie database.

    :return: A list of class expanding Document type
    """
    models = set()

    if isinstance(root_module, str):
        root_module = sys.modules[root_module]

    for package in pkgutil.walk_packages(
        root_module.__path__, root_module.__name__ + "."
    ):
        module = importlib.import_module(package.name)

        for name, cls in inspect.getmembers(module, inspect.isclass):
            if inspect.isabstract(cls):
                continue
            if not issubclass(cls, Document):
                continue
            if cls in [Document, MightstoneDocument, MightstoneSerializableDocument]:
                continue

            patch_beanie_document(cls)
            models.add(cls)

    return list(models)


def patch_beanie_document(model: Type[Document]):
    """
    Beanie documents require a Settings inner class that defines its configuration.
    We forcefully patch the models globally to add:

    * bson_encoders support for:
        - datetime.date
        - pydantic Color
    * name the model collection from the module if not already defined

    :param model:
    :return:
    """

    if hasattr(model, "Settings") and hasattr(model.Settings, "name"):
        collection_name = model.Settings.name
    else:
        notable_parent_package = [
            pkg
            for pkg in model.__module__.split(".")
            if pkg not in ["models", "services"]
        ]
        collection_name = "_".join(
            notable_parent_package
            + [model.__name__.lower().replace("serializable", "")]
        )
        if collection_name[-1] != "s":
            collection_name += "s"

    model.Settings = type(
        "Settings",
        (object,),
        {
            "bson_encoders": {
                datetime.date: lambda dt: datetime.datetime(
                    year=dt.year,
                    month=dt.month,
                    day=dt.day,
                    hour=0,
                    minute=0,
                    second=0,
                ),
                Color: lambda c: c.as_hex(),
            },
            "name": collection_name,
        },
    )  # type: ignore


class MightstoneError(Exception):
    pass



import asyncio

import click

from mightstone.cli.models import MightstoneCli, pass_mightstone
from mightstone.cli.utils import pretty_print
from mightstone.services.mtgjson import MtgJsonCompression


@click.group()
@pass_mightstone
@click.option(
    "--compression",
    type=click.Choice([t for t in MtgJsonCompression]),
    default=MtgJsonCompression.GZIP,
)
def mtgjson(cli: MightstoneCli, compression):
    cli.app.mtg_json.set_compression(compression)


@mtgjson.command()
@pass_mightstone
def meta(cli: MightstoneCli):
    """Display the current version."""
    pretty_print(asyncio.run(cli.app.mtg_json.meta_async()), cli.format)


@mtgjson.command()
@pass_mightstone
def card_types(cli: MightstoneCli):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(cli.app.mtg_json.card_types_async()), cli.format)


@mtgjson.command()
@pass_mightstone
@click.argument("code", type=str)
def set(cli: MightstoneCli, **kwargs):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(cli.app.mtg_json.set_async(**kwargs)), cli.format)


@mtgjson.command()
@pass_mightstone
def compiled_list(cli: MightstoneCli):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(cli.app.mtg_json.compiled_list_async()), cli.format)


@mtgjson.command()
@pass_mightstone
def keywords(cli: MightstoneCli):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(cli.app.mtg_json.keywords_async()), cli.format)


@mtgjson.command()
@pass_mightstone
def enum_values(cli: MightstoneCli):
    """Display every card type of any type of card."""
    pretty_print(asyncio.run(cli.app.mtg_json.enum_values_async()), cli.format)



import html
import re
from typing import Optional, Union

from mightstone.core import MightstoneModel


class WikiTag(MightstoneModel):
    tag: str
    extra: Optional["WikiString"] = None
    kwargs: dict[str, "WikiString"] = {}
    args: list["WikiString"] = []

    def get_kwarg(self, keys: list[str]) -> Union[str, None]:
        for key in keys:
            if key in self.kwargs:
                return str(self.kwargs[key])

        return None

    def get_arg(self, index: int) -> Union[str, None]:
        try:
            return str(self.args[index])
        except IndexError:
            return None

    def __str__(self):
        parts = [self.tag]
        if self.extra:
            parts = [f"{self.tag} {self.extra}"]

        if self.args:
            parts.append(" | ".join(map(str, self.args)))

        if self.kwargs:
            parts.append(
                " | ".join(map(lambda kv: f"{kv[0]} = {kv[1]}", self.kwargs.items()))
            )

        return "{{ " + " | ".join(parts) + " }}"


WIKILINK = re.compile(r"\[([^]]+)]")


class WikiString(MightstoneModel):
    tokens: list[Union[str, WikiTag]] = list()

    def __str__(self):
        return " ".join(map(WikiString.escape_part, self.tokens))

    @staticmethod
    def escape_part(part) -> str:
        if isinstance(part, str):
            return html.unescape(WIKILINK.sub(r"\1", part))
        return str(part)

    @staticmethod
    def from_string(string: str):
        return WikiString(tokens=[string])



import importlib
from pathlib import Path

import click
from pydantic_core._pydantic_core import Url

from mightstone.cli.models import MightstoneCli, pass_mightstone


@click.group()
def wiki():
    pass


@wiki.group()
def scrape(): ...


module = importlib.import_module("mightstone.rule.data")
default_output = Path(module.__path__[0]).joinpath("abilities.json")


@scrape.command()
@click.argument("output", type=click.File("w"), default=default_output)
@pass_mightstone
def abilities(mightstone: MightstoneCli, output: click.utils.LazyFile):
    print("Building abilities from mtg wiki...")
    abilities = mightstone.app.wiki.scrape_abilities()

    print("Saving informations...")
    output.write(abilities.model_dump_json(indent=2))
    print(f"{len(abilities.abilities)} abilities saved into {output.name}")



import click

from mightstone.ass import aiterator_to_list
from mightstone.cli.models import MightstoneCli, pass_mightstone
from mightstone.cli.utils import pretty_print
from mightstone.services.edhrec.api import (
    EdhRecCategory,
    EdhRecIdentity,
    EdhRecPeriod,
    EdhRecType,
)


@click.group()
def edhrec():
    pass


@edhrec.command()
@pass_mightstone
@click.argument("name", nargs=1)
@click.argument("sub", required=False)
def commander(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.edhrec_static.commander(**kwargs), cli.format)


@edhrec.command()
@pass_mightstone
@click.argument("identity", required=False)
@click.option("-l", "--limit", type=int)
def tribes(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.tribes_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.argument("identity", required=False)
@click.option("-l", "--limit", type=int)
def themes(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.themes_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-l", "--limit", type=int)
def sets(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.sets_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-l", "--limit", type=int)
def companions(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.companions_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-i", "--identity", type=str)
@click.option("-l", "--limit", type=int)
def partners(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.partners_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-i", "--identity", type=str)
@click.option("-l", "--limit", type=int, default=100)
def commanders(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.commanders_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.argument("identity", type=click.Choice([t.value for t in EdhRecIdentity]))
@click.option("-l", "--limit", type=int, default=100)
def combos(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.combos_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.argument("identity", type=click.Choice([t.value for t in EdhRecIdentity]))
@click.argument("identifier", type=str)
@click.option("-l", "--limit", type=int, default=100)
def combo(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.combo_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.argument("year", required=False, type=int)
@click.option("-l", "--limit", type=int)
def salt(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.salt_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-t", "--type", type=click.Choice([t.value for t in EdhRecType]))
@click.option("-p", "--period", type=click.Choice([t.value for t in EdhRecPeriod]))
@click.option("-l", "--limit", type=int)
def top_cards(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.top_cards_async(**kwargs)), cli.format
    )


@edhrec.command()
@pass_mightstone
@click.option("-c", "--category", type=click.Choice([t.value for t in EdhRecCategory]))
@click.option("-t", "--theme", type=str)
@click.option("--commander", type=str)
@click.option("-i", "--identity", type=str)
@click.option("-s", "--set", type=str)
@click.option("-l", "--limit", type=int)
def cards(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.edhrec_static.cards_async(**kwargs)), cli.format
    )



import asyncio
import logging
from functools import wraps
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Generator,
    List,
    Optional,
    TypeVar,
)

import asyncstdlib
from asgiref.sync import async_to_sync

logger = logging.getLogger(__name__)

T = TypeVar("T")


@async_to_sync
async def aiterator_to_list(agen: AsyncGenerator[T, Any], limit=100) -> List[T]:
    """
    Transforms an async iterator into a sync list

    :param agen: Asynchronous iterator
    :param limit: Max item to return
    :return: The list of items
    """

    if limit:
        ait = asyncstdlib.islice(agen.__aiter__(), limit)
    else:
        ait = agen.__aiter__()
    return [item async for item in ait]


R = TypeVar("R")


def synchronize(
    f: Callable[..., Coroutine[Any, Any, R]],
    docstring: Optional[str] = None,
) -> Callable[..., R]:
    qname = f"{f.__module__}.{f.__qualname__}"

    @wraps(f)
    def inner(*args, **kwargs) -> R:
        executor = async_to_sync(f)
        return executor(*args, **kwargs)

    if docstring:
        inner.__doc__ = docstring
    else:
        inner.__doc__ = (
            f"Sync version of :func:`~{qname}`, same behavior but "
            "wrapped by :func:`~asgiref.sync.async_to_sync`."
        )

    return inner


def sync_generator(
    f: Callable[..., AsyncGenerator[R, None]],
    docstring: Optional[str] = None,
) -> Callable[..., Generator[R, None, None]]:
    qname = f"{f.__module__}.{f.__qualname__}"
    loop = asyncio.get_event_loop()

    @wraps(f)
    def inner(*args, **kwargs) -> Generator[R, None, None]:
        async_generator = f(*args, **kwargs)
        try:
            while True:
                yield loop.run_until_complete(async_generator.__anext__())
        except StopAsyncIteration:
            return

    if docstring:
        inner.__doc__ = docstring
    else:
        inner.__doc__ = (
            f"Sync version of :func:`~{qname}`, same behavior but "
            "wrapped by :func:`~asgiref.sync.async_to_sync`."
        )

    return inner



import asyncio
import logging
import re
from datetime import date, timedelta
from io import StringIO
from typing import List, Optional

from mightstone.ass import synchronize
from mightstone.services import MightstoneHttpClient
from mightstone.services.wotc.models import (
    ComprehensiveRules,
    SerializableComprehensiveRules,
)

logger = logging.getLogger("mightstone")


class RuleExplorer(MightstoneHttpClient):
    async def open_async(self, path: Optional[str] = None) -> ComprehensiveRules:
        """
        Open a local or remote comprehensive rule document, if no path is provided
        then the latest rules from Wizards of the Coast website is pulled.

        :param path: A local path, or an URL to the rule document
        :return: A ``ComprehensiveRules`` instance
        """
        if not path:
            path = await self.latest_async()

        if path.startswith("http"):
            f = await self.client.get(path)
            f.raise_for_status()
            try:
                content = StringIO(f.content.decode("UTF-8"))
            except UnicodeDecodeError:
                content = StringIO(f.content.decode("iso-8859-1"))

            return ComprehensiveRules.parse(content)

        with open(path, "r") as f:
            return ComprehensiveRules.parse(f)

    open = synchronize(open_async)

    async def latest_async(self) -> str:
        """
        Resolves wizard latest published ruleset

        :return: The url of the latest ruleset to date
        """
        f = await self.client.get("https://magic.wizards.com/en/rules")
        f.raise_for_status()
        return self.match_text_last_url(f.text)

    latest = synchronize(latest_async)

    async def explore_async(
        self, after: date, before: Optional[date] = None, concurrency=3
    ) -> List[str]:
        """
        Explore the wizards of the coast website to find any rule between two timestamp.

        Wizards don’t support an historic index of previous rules, this method tries to
        compensate by providing a brute force attempt to pull magic rule history.
        This method will brute force every possible rule using the current format:

        https://media.wizards.com/YYYY/downloads/MagicComp%20Rules%20{YYYY}{MM}{DD}.txt

        :param after: The min date to scan
        :param before: The max date to scan (defaults to today)
        :param concurrency: The max number of concurrent HTTP requests
        :return: A list of existing rules url
        """
        if not before:
            before = date.today()

        urls = []
        for n in range(int((before - after).days)):
            d = after + timedelta(n)
            urls.append(d.strftime("/%Y/downloads/MagicComp%%20Rules%%20%Y%m%d.txt"))
            urls.append(d.strftime("/%Y/downloads/MagicCompRules%%20%Y%m%d.txt"))

        found = []
        sem = asyncio.Semaphore(concurrency)

        async def test_url(url: str):
            async with sem:
                logger.debug("GET %s", url)
                resp = await self.client.get(url)
                if resp.is_success:
                    logger.info("Found %s", url)
                    found.append(url)

        tasks = []
        for url in urls:
            task = asyncio.ensure_future(test_url(url))
            tasks.append(task)
        await asyncio.gather(*tasks, return_exceptions=True)

        return found

    explore = synchronize(explore_async)

    LAST_TEXT_URL_PATTERN = re.compile(r"https://.+MagicCompRules.+\.txt")

    @classmethod
    def match_text_last_url(cls, html_source: str):
        res = cls.LAST_TEXT_URL_PATTERN.search(html_source)
        if res:
            return res.group(0).replace(" ", "%20")

        raise RuntimeError("Unable to find URL of the last comprehensive rules")



from mightstone.rule.models.color import Color, ColorPie

White = Color(symbol="w", index=0)
Blue = Color(symbol="u", index=1)
Black = Color(symbol="b", index=2)
Red = Color(symbol="r", index=3)
Green = Color(symbol="g", index=4)

all_colors = [White, Blue, Black, Red, Green]

color_pie = ColorPie([White, Blue, Black, Red, Green])

identities = color_pie.build_identity_map()
identities.refine("", "Colorless")
identities.refine("w", "Mono-White", ["White"])
identities.refine("u", "Mono-Blue", ["Blue"])
identities.refine("b", "Mono-Black", ["Black"])
identities.refine("r", "Mono-Red", ["Red"])
identities.refine("g", "Mono-Green", ["Green"])
identities.refine("wu", "Azorius", ["Azorius Senate"])
identities.refine("ub", "Dimir", ["House Dimir"])
identities.refine("br", "Rakdos", ["Cult of Rakdos"])
identities.refine("rg", "Gruul", ["Gruul Clans"])
identities.refine("gu", "Simic", ["Simic Combine", "Quandrix"])
identities.refine(
    "gw",
    "Selesnya",
    ["Selesnya Conclave"],
)
identities.refine("wb", "Orzhov", ["Orzhov Syndicate", "Silverquill"])
identities.refine("ur", "Izzet", ["Izzet League", "Prismari"])
identities.refine("rw", "Boros", ["Boros Legion", "Lorehold"])
identities.refine("gb", "Golgari", ["Golgari Swarm", "Witherbloom"])
identities.refine("wub", "Esper", ["Obscura"])
identities.refine("ubr", "Grixis", ["Maestros"])
identities.refine("brg", "Jund", ["Riveteers"])
identities.refine("rgw", "Naya", ["Cabaretti"])
identities.refine("guw", "Bant", ["Brokers"])
identities.refine("wbg", "Abzan", ["Abzan Houses", "Indatha"])
identities.refine("urw", "Jeskai", ["Jeskai Way", "Raugrin"])
identities.refine("bgu", "Sultai", ["Sultai Brood", "Zagoth"])
identities.refine("rwb", "Mardu", ["Mardu Horde", "Savai"])
identities.refine("gur", "Temur", ["Temur Frontier", "Ketria"])
identities.refine("wubr", "Artifice", ["Yore-Tiller", "Green-less", "Yore"])
identities.refine("ubrg", "Chaos", ["Glint-Eye", "White-less", "Glint"])
identities.refine("brgw", "Aggression", ["Dune-Brood", "Blue-less", "Dune"])
identities.refine("rgwu", "Altruism", ["Ink-Treader", "Black-less", "Ink"])
identities.refine("gwub", "Growth", ["Witch-Maw", "Red-less", "Witch"])
identities.refine("wubrg", "Five-Color", ["5c", "Rainbow"])



from pathlib import Path

from ..models.ability import AbilityList

"""
The MTG ruleset
"""

abilities = AbilityList.model_validate_json(
    Path(__file__).parent.joinpath("abilities.json").read_bytes()
)



import logging
import urllib.parse
from typing import Generator, Union

from bs4 import BeautifulSoup, Tag
from pydantic_core._pydantic_core import ValidationError
from pyparsing import (
    CharsNotIn,
    Combine,
    Forward,
    OneOrMore,
    Optional,
    ParseResults,
    Suppress,
    Word,
    alphanums,
    alphas,
    delimited_list,
    nested_expr,
)

from mightstone.ass import synchronize
from mightstone.rule.models.ability import Ability, AbilityList, AbilityType
from mightstone.services import MightstoneHttpClient

from ...rule.models.color import Identity, IdentityMap
from .models import WikiString, WikiTag


class Wiki(MightstoneHttpClient):
    """
    Scryfall API client
    """

    base_url = "https://mtg.fandom.com"

    async def export_pages_async(self, pages: list[str]):
        response = await self.client.post(
            "/wiki/Special:Export",
            data={
                "catname": "",
                "pages": "\n".join(pages),
                "curonly": "1",
                "wpDownload": 1,
                "wpEditToken": "+\\",
                "title": "Special:Export",
            },
        )
        return response.content

    export_pages = synchronize(export_pages_async)

    async def export_category_async(self, category) -> bytes:
        response = await self.client.post(
            "/wiki/Special:Export",
            data={
                "addcat": "Add",
                "catname": category,
                "pages": "",
                "curonly": "1",
                "wpDownload": 1,
                "wpEditToken": "+\\",
                "title": "Special:Export",
            },
        )
        soup = BeautifulSoup(response.content, "lxml")
        pages_as_source = soup.find("textarea", {"name": "pages"})
        if not isinstance(pages_as_source, Tag):
            raise RuntimeError("Unable to resolve pages for category %s" % category)
        pages = [page.strip() for page in pages_as_source.decode_contents().split("\n")]
        return await self.export_pages_async(pages)

    export_category = synchronize(export_category_async)

    async def all_pages(self):
        raise NotImplementedError()

    async def scrape_abilities_async(self) -> AbilityList:
        from ...rule.data.color import identities

        categories = [
            "Keywords/Static",
            "Keywords/Activated",
            "Keywords/Characteristic-defining",
            "Keywords/Evasion",
            "Keywords/Spell",
            "Keywords/Triggered",
        ]

        list = AbilityList()
        for category in categories:
            export = await self.export_category_async(category)
            adapter = WikiAbilityAdapter(self.base_url, export, identities=identities)
            for ability in adapter.abilities():
                list.abilities.append(ability)
        return list

    scrape_abilities = synchronize(scrape_abilities_async)


class WikiParser:
    LBRACE = Suppress("{{")
    RBRACE = Suppress("}}")
    EQUALS = Suppress("=")
    PIPE = Suppress("|")

    WIKI_TAG = Forward()
    STRING_SEQUENCE = Combine(
        OneOrMore(~LBRACE + ~RBRACE + ~PIPE + CharsNotIn("\n", exact=1))
    )
    TEMPLATE_STRING = OneOrMore(WIKI_TAG | STRING_SEQUENCE)
    PROPERTY_NAME = Word(alphanums + "_")
    WIKI_BLOCK_NAME = Word(alphas + alphanums + "-_+").set_results_name("wiki_tag_name")

    PROPERTY_AS_KWARGS = (PROPERTY_NAME + EQUALS + TEMPLATE_STRING).set_results_name(
        "kwargs", list_all_matches=True
    )
    PROPERTY_AS_ARG = TEMPLATE_STRING.set_results_name("args", list_all_matches=True)
    PROPERTY = PROPERTY_AS_KWARGS | PROPERTY_AS_ARG
    PROPERTIES = delimited_list(PROPERTY, delim=PIPE)

    WIKI_TAG << nested_expr(
        LBRACE,
        RBRACE,
        content=WIKI_BLOCK_NAME
        + Optional(TEMPLATE_STRING.set_results_name("extra"))
        + Optional(PIPE + PROPERTIES)
        + Optional(PIPE),
    ).set_results_name("wiki_tag")

    def __init__(self, name, content):
        self.name = name
        self.content = content

    @classmethod
    def flatten_results(
        cls,
        results: ParseResults,
    ) -> Generator[ParseResults, None, None]:
        for result in results:
            if isinstance(result, str):
                continue

            if result.wiki_tag_name:
                yield result

            yield from cls.flatten_results(result)
        return

    def get_infobox(self) -> Union[WikiTag, None]:
        iterator = self.get_wiki_tags(["Infobox"])
        return next(iterator, None)

    def get_stats(self) -> Generator[WikiTag, None, None]:
        found = False
        for tag in self.get_wiki_tags(["stats"], recurse=True):
            found = True
            yield tag

        if not found:
            yield WikiTag(tag="stats")

    def extract_glossary_from_rule(self, tag: WikiTag) -> Union[WikiTag, str, None]:
        if tag.tag == "CR+G":
            if not len(tag.args):
                return self.name
            return tag.args[0].tokens[0]
        else:
            if "lookup" in tag.kwargs:
                return tag.kwargs["lookup"].tokens[0]
            if len(tag.args) > 1 and tag.args[0].tokens[0] == "glossary":
                return tag.args[1].tokens[0]
        return None

    def get_rules(self) -> Generator[Union[str, WikiTag], None, None]:
        for tag in self.get_wiki_tags(["CR"]):
            if not self.extract_glossary_from_rule(tag):
                yield tag.args[0].tokens[0]

    def get_glossaries(self) -> Generator[Union[str, WikiTag], None, None]:
        for tag in self.get_wiki_tags(["CR+G", "CR"]):
            glossary = self.extract_glossary_from_rule(tag)
            if glossary:
                yield glossary

    def get_wiki_tags(
        self, tags: list[str], max_matches=100, recurse=False
    ) -> Generator[WikiTag, None, None]:
        matches = self.WIKI_TAG.search_string(self.content, max_matches)

        if recurse:
            iterator = self.flatten_results(matches)
        else:
            iterator = [tag for match in matches for tag in match.wiki_tag]  # type: ignore

        for tag in iterator:
            if tag.wiki_tag_name in tags:
                yield self.build_wiki_tag(tag)
        return

    def build_wiki_string(self, tokens: list[Union[ParseResults, str]]) -> WikiString:
        out = WikiString()
        for token in tokens:
            if isinstance(token, str):
                out.tokens.append(token.strip())
            else:
                out.tokens.append(self.build_wiki_tag(token))
        return out

    def build_wiki_tag(self, wiki_tag: ParseResults) -> WikiTag:
        args = []
        if wiki_tag.args:
            args = [self.build_wiki_string(arg) for arg in wiki_tag.args]

        kwargs = {}
        if wiki_tag.kwargs:
            for x in wiki_tag.kwargs:
                kwargs[x[0].strip()] = self.build_wiki_string(x[1:])

        extra = None
        if wiki_tag.extra:
            extra = self.build_wiki_string(wiki_tag.extra)

        return WikiTag(
            tag=wiki_tag.wiki_tag_name, args=args, kwargs=kwargs, extra=extra
        )


class WikiAbilityAdapter:
    def __init__(self, base_url: str, content: bytes, identities: IdentityMap):
        self.base_url = base_url
        self.soup = BeautifulSoup(content, "lxml")
        self.identities = identities

    def pages(self) -> list[Tag]:
        return self.soup.find_all("page") or []

    @staticmethod
    def map_ability_types(infobox: WikiTag) -> list[AbilityType]:
        types: list[str] = [
            infobox.get_kwarg([key]) or ""
            for key in sorted(infobox.kwargs.keys())
            if key.startswith("type")
        ]

        out = []
        for t in types:
            try:
                out.append(AbilityType(t.lower()))
            except ValueError:
                ...
        return out

    def abilities(self) -> Generator[Ability, None, None]:
        for page in self.pages():
            title_as_str = page.find_next("title").text  # type: ignore
            parser = WikiParser(
                title_as_str,
                page.find_next("text").text,  # type: ignore
            )
            url = f"{self.base_url}/wiki/{urllib.parse.quote(title_as_str)}"

            infobox = parser.get_infobox()
            if not infobox:
                logging.warning("Unable to parse %s, no infobox found" % (url))
                continue
            rules = list(parser.get_rules())
            stats = list(parser.get_stats())

            for stat in stats:
                try:
                    yield Ability.model_validate(
                        {
                            "name": stat.get_arg(0) or title_as_str,
                            "types": self.map_ability_types(infobox),
                            "rules": rules,
                            "wiki": url,
                            "introduced": infobox.get_kwarg(["first"]),
                            "last_seen": infobox.get_kwarg(["last"]),
                            "has_cost": bool(infobox.get_kwarg(["cost", "N"])),
                            "reminder": infobox.get_kwarg(["reminder"]),
                            "stats": self.map_stats(stat),
                            "storm": infobox.get_kwarg(["storm"]),
                        }
                    )
                except ValidationError as e:
                    logging.warning(
                        "%s (%s) failed to be validated: %s"
                        % (url, stat.get_arg(0) or title_as_str, e)
                    )
                    continue

    def map_stats(self, tag: WikiTag) -> dict[str, int]:
        out = {}
        for k, v in tag.kwargs.items():
            try:
                identity = self.identities[k.lower()]
                out[identity.canonical] = int(v.tokens[0])  # type: ignore
            except KeyError:
                ...

        return out



import logging
from enum import Enum
from typing import AsyncGenerator, List, Optional, Tuple, Union

import asyncstdlib
from httpx import HTTPStatusError
from pydantic import ValidationError

from mightstone.ass import sync_generator, synchronize
from mightstone.services import MightstoneHttpClient, ServiceError
from mightstone.services.edhrec.models import (
    EdhRecCardItem,
    EdhRecCategory,
    EdhRecCommander,
    EdhRecFilterQuery,
    EdhRecPeriod,
    EdhRecRecs,
    slugify,
)

logger = logging.getLogger("mightstone")


class EdhRecIdentity(str, Enum):
    COLORLESS = "colorless"
    W = "w"
    U = "u"
    B = "b"
    R = "r"
    G = "g"
    WU = "wu"
    UB = "ub"
    BR = "br"
    RG = "rg"
    GW = "gw"
    WB = "wb"
    UR = "ur"
    BG = "bg"
    RW = "rw"
    GU = "gu"
    WUB = "wub"
    UBR = "ubr"
    BRG = "brg"
    RGW = "rgw"
    GWU = "gwu"
    WBG = "wbg"
    URW = "urw"
    BGU = "bgu"
    RWB = "rwb"
    GUR = "gur"
    WUBR = "wubr"
    UBRG = "ubrg"
    BRGW = "brgw"
    RGWU = "rgwu"
    GWUB = "gwub"
    WUBRG = "wubrg"


class EdhRecTag(str, Enum):
    TRIBES = "tribes"
    SET = "sets"
    NONE = ""
    THEME_POPULARITY = "themesbypopularitysort"
    THEME = "themes"
    COMMANDER = "topcommanders"
    COMPANION = "companions"


class EdhRecType(str, Enum):
    CREATURE = "creatures"
    INSTANT = "instants"
    SORCERY = "sorceries"
    ARTIFACT = "artifacts"
    ARTIFACT_EQUIPMENT = "equipment"
    ARTIFACT_UTILITY = "utility-artifacts"
    ARTIFACT_MANA = "mana-artifacts"
    ENCHANTMENT = "enchantments"
    ENCHANTMENT_AURA = "auras"
    PLANESWALKER = "planeswalker"
    LAND = "lands"
    LAND_UTILITY = "utility-lands"
    LAND_FIXING = "color-fixing-lands"


class EdhRecApi(MightstoneHttpClient):
    """
    HTTP client for dynamic data hosted at https://edhrec.com/api/
    """

    base_url = "https://edhrec.com"

    async def recommendations_async(self, commanders: List[str], cards: List[str]):
        """
        Obtain EDHREC recommendations for a given commander (or partners duo)
        for a given set of cards in the deck.

        Returns a list of 99 suggested cards not contained in the list
        :param commanders: A list of one or two commander card name
        :param cards: A list of card name
        :exception ClientResponseError
        :returns An EdhRecRecs object
        """
        try:
            session = self.client
            async with session.post(
                "/api/recs/",
                json={"cards": cards, "commanders": commanders},
            ) as f:
                f.raise_for_status()
                data = await f.json()

                if data.get("errors"):
                    raise ServiceError(
                        message=data.get("errors")[0],
                        data=data,
                        url=f.request_info.real_url,
                        status=f.status,
                    )

                return EdhRecRecs.model_validate(data)

        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status_code,
            )

    recommendations = synchronize(recommendations_async)

    async def filter_async(
        self, commander: str, query: EdhRecFilterQuery
    ) -> EdhRecCommander:
        """
        Read Commander related information, and return an EdhRecCommander object

        :param commander: Commander name or slug
        :param query: An EdhRecFilterQuery object describing the request
        :return: An EdhRecCommander representing answer
        """
        try:
            session = await self.client
            async with session.get(
                "/api/filters/",
                params={
                    "f": str(query),
                    "dir": "commanders",
                    "cmdr": slugify(commander),
                },
            ) as f:
                f.raise_for_status()
                return EdhRecCommander.parse_payload(await f.json())

        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status_code,
            )

    filter = synchronize(filter_async)


class EdhRecStatic(MightstoneHttpClient):
    """
    HTTP client for static JSON data hosted at https://json.edhrec.com
    """

    base_url = "https://json.edhrec.com"

    async def commander_async(
        self, name: str, sub: Optional[str] = None
    ) -> EdhRecCommander:
        """

        :param name: Commander
        :param sub:
        :return:
        """
        path = f"commanders/{slugify(name)}.json"
        if sub:
            path = f"commanders/{slugify(name)}/{slugify(sub)}.json"

        data = await self._get_static_page(path)

        return EdhRecCommander.parse_payload(data)

    commander = synchronize(commander_async)

    async def tribes_async(
        self,
        identity: Optional[Union[EdhRecIdentity, str]] = None,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if identity:
            identity = EdhRecIdentity(identity)
            async for item in self._page_item_generator(
                f"commanders/{identity.value}.json",
                EdhRecTag.TRIBES,
                related=True,
                limit=limit,
            ):
                yield item
        else:
            async for item in self._page_item_generator(
                "tribes.json", EdhRecTag.TRIBES, limit=limit
            ):
                yield item

    tribes = sync_generator(tribes_async)

    async def themes_async(
        self,
        identity: Optional[Union[EdhRecIdentity, str]] = None,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if identity:
            identity = EdhRecIdentity(identity)
            async for item in self._page_item_generator(
                f"commanders/{identity.value}.json",
                EdhRecTag.THEME,
                related=True,
                limit=limit,
            ):
                yield item
        else:
            async for item in self._page_item_generator(
                "themes.json", EdhRecTag.THEME_POPULARITY, limit=limit
            ):
                yield item

    themes = sync_generator(themes_async)

    async def sets_async(
        self, limit: Optional[int] = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        async for item in self._page_item_generator(
            "sets.json", EdhRecTag.SET, limit=limit
        ):
            yield item

    sets = sync_generator(sets_async)

    async def salt_async(
        self, year: Optional[int] = None, limit: Optional[int] = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "top/salt.json"
        if year:
            path = f"top/salt-{year}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    salt = sync_generator(salt_async)

    async def top_cards_async(
        self,
        type: Optional[EdhRecType] = None,
        period: EdhRecPeriod = EdhRecPeriod.PAST_WEEK,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        period = EdhRecPeriod(period)
        if type:
            type = EdhRecType(type)
            async for item in self._page_item_generator(
                f"top/{type.value}.json", period, limit=limit
            ):
                yield item
            return

        if period == EdhRecPeriod.PAST_WEEK:
            path = "top/week.json"
        elif period == EdhRecPeriod.PAST_MONTH:
            path = "top/month.json"
        else:
            path = "top/year.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    top_cards = sync_generator(top_cards_async)

    async def cards_async(
        self,
        theme: Optional[str] = None,
        commander: Optional[str] = None,
        identity: Optional[Union[EdhRecIdentity, str]] = None,
        set: Optional[str] = None,
        category: Optional[EdhRecCategory] = None,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        if category:
            category = EdhRecCategory(category)

        if not theme and not commander and not set:
            raise ValueError("You must either provide a theme, commander or set")

        if commander:
            if theme:
                raise ValueError("commander and theme options are mutually exclusive")
            if identity:
                raise ValueError(
                    "commander and identity options are mutually exclusive"
                )
            if set:
                raise ValueError("commander and set options are mutually exclusive")

            slug = slugify(commander)
            path = f"commanders/{slug}.json"
            if theme:
                path = f"commanders/{slug}/{slugify(theme)}.json"
            async for item in self._page_item_generator(path, category, limit=limit):
                yield item

            return

        if set:
            if theme:
                raise ValueError("set and theme options are mutually exclusive")
            if identity:
                raise ValueError("set and identity options are mutually exclusive")
            async for item in self._page_item_generator(
                f"sets/{slugify(set)}.json", category, limit=limit
            ):
                yield item
            return

        if identity and not theme:
            raise ValueError("you must specify a theme to search by color identity")

        path = f"themes/{slugify(theme)}.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"themes/{slugify(theme)}/{identity.value}.json"
        async for item in self._page_item_generator(path, category, limit=limit):
            yield item

    cards = sync_generator(cards_async)

    async def companions_async(
        self, limit: Optional[int] = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        async for item in self._page_item_generator(
            "companions.json", EdhRecTag.COMPANION, limit=limit
        ):
            yield item

    companions = sync_generator(companions_async)

    async def partners_async(
        self,
        identity: Optional[Union[EdhRecIdentity, str]] = None,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "partners.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"partners/{identity.value}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    partners = sync_generator(partners_async)

    async def commanders_async(
        self,
        identity: Optional[Union[EdhRecIdentity, str]] = None,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        path = "commanders.json"
        if identity:
            identity = EdhRecIdentity(identity)
            path = f"commanders/{identity.value}.json"
        async for item in self._page_item_generator(path, limit=limit):
            yield item

    commanders = sync_generator(commanders_async)

    async def combos_async(
        self, identity: Union[EdhRecIdentity, str], limit: Optional[int] = None
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        identity = EdhRecIdentity(identity)
        async for item in self._page_item_generator(
            f"combos/{identity.value}.json", limit=limit
        ):
            yield item

    combos = sync_generator(combos_async)

    async def combo_async(
        self,
        identity: str,
        identifier: Union[EdhRecIdentity, str],
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        identity = EdhRecIdentity(identity)
        async for item in self._page_item_generator(
            f"combos/{identity.value}/{int(identifier)}.json", limit=limit
        ):
            yield item

    combo = sync_generator(combo_async)

    async def _page_item_generator(
        self,
        path,
        tag: Optional[
            Union[EdhRecTag, EdhRecType, EdhRecPeriod, EdhRecCategory]
        ] = None,
        related=False,
        limit: Optional[int] = None,
    ) -> AsyncGenerator[EdhRecCardItem, None]:
        """
        Async generator that will wrap Pydantic validation
        and ensure that no validation error are raised
        """
        ttag: Optional[str] = None
        if tag:
            ttag = tag.value

        enumerator = asyncstdlib.enumerate(self._get_page(path, ttag, related))
        async with asyncstdlib.scoped_iter(enumerator) as protected_enumerator:
            async for i, (ttag, page, index, item) in protected_enumerator:
                if limit and i == limit:
                    logger.debug(f"Reached limit of {limit}, stopping iteration")
                    return

                try:
                    yield EdhRecCardItem.parse_payload(item, ttag)
                except ValidationError as e:
                    logger.warning(
                        "Failed to parse an EDHREC item from %s at page %d, index %d",
                        path,
                        page,
                        index,
                    )
                    logger.debug(e.json())

    async def _get_static_page(self, path) -> dict:
        try:
            f = await self.client.get(f"/pages/{path}")
            f.raise_for_status()
            return f.json()
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from EDHREC",
                url=e.request.url,
                status=e.response.status_code,
            )

    async def _get_page(
        self, path, tag: Optional[str] = None, related=False
    ) -> AsyncGenerator[Tuple[str, int, int, dict], None]:
        """
        Read a EDHREC page data, and return it as a tuple:
        - tag as string
        - page
        - index
        - the payload itself
        """
        data = await self._get_static_page(path)
        page = 1

        if related:
            iterator = [
                {"tag": tag, "cardviews": data.get("relatedinfo", {}).get(tag, [])}
            ]
        else:
            iterator = (
                data.get("container", {}).get("json_dict", {}).get("cardlists", [])
            )

        for item_list in iterator:
            current_tag = item_list.get("tag", "")
            if tag is not None and str(tag) != current_tag:
                continue

            for index, item in enumerate(item_list.get("cardviews", [])):
                yield (
                    current_tag,
                    page,
                    index,
                    item,
                )

            while item_list.get("more"):
                item_list = await self._get_static_page(f"{item_list.get('more')}")
                page += 1
                for index, item in enumerate(item_list.get("cardviews", [])):
                    yield current_tag, page, index, item



import re
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import Field

from mightstone.core import MightstoneModel

salt_parser = re.compile(r"Salt Score: (?P<salt>[\d.]+)\n")
synergy_parser = re.compile(r"(?P<synergy>[\d.]+)% synergy")


class EdhRecCategory(Enum):
    TOP_COMMANDER = "topcommanders"
    COMMANDER = "commanders"
    NEW = "newcards"
    HIGH_SYNERGY = "highsynergycards"
    TOP_CARD = "topcards"
    CREATURE = "creatures"
    INSTANT = "instants"
    SORCERY = "sorceries"
    ARTEFACT_UTIL = "utilityartifacts"
    ENCHANTMENT = "enchantments"
    PLANEWALKER = "planeswalkers"
    LAND_UTIL = "utilitylands"
    ARTEFACT_MANA = "manaartifacts"
    LAND = "lands"


class EdhRecPeriod(Enum):
    PAST_WEEK = "pastweek"
    PAST_MONTH = "pastmonth"
    PAST_2YEAR = "past2years"


class EdhRecCardRef(MightstoneModel):
    name: str
    url: str


class EdhRecCard(MightstoneModel):
    cmc: int
    color_identity: List[str]

    combos: Optional[bool] = None
    label: Optional[str] = None
    legal_commander: Optional[bool] = None

    image_uris: List[dict]
    is_commander: Optional[bool] = None
    layout: str
    name: str
    names: List[str]
    inclusion: Optional[int] = None
    num_decks: Optional[int] = None
    potential_decks: Optional[int] = None
    precon: Optional[str] = None
    prices: dict
    primary_type: str
    rarity: str
    salt: float
    sanitized: str
    sanitized_wo: str
    type: str
    url: Optional[str] = None
    aetherhub_uri: Optional[str] = None
    archidekt_uri: Optional[str] = None
    deckstats_uri: Optional[str] = None
    moxfield_uri: Optional[str] = None
    mtggoldfish_uri: Optional[str] = None
    scryfall_uri: Optional[str] = None
    spellbook_uri: Optional[str] = None


class EdhRecCommanderSub(MightstoneModel):
    count: int
    suffix: str = Field(alias="href-suffix")
    value: str


class EdhRecCommanderDistribution(MightstoneModel):
    artifact: int = 0
    creature: int = 0
    enchantment: int = 0
    instant: int = 0
    land: int = 0
    planeswalker: int = 0
    sorcery: int = 0


class EdhRecCardItem(MightstoneModel):
    tag: str
    name: str
    slug: str
    url: Path
    label: Optional[str] = None
    inclusion: Optional[int] = None
    cards: Optional[List[EdhRecCardRef]] = None
    count: Optional[int] = None
    num_decks: Optional[int] = None
    potential_decks: Optional[int] = None
    synergy: Optional[float] = None
    salt: Optional[float] = None

    @classmethod
    def parse_payload(cls, data: dict, tag: Optional[str]):
        salt = salt_parser.search(data.get("label", "unspecified"))
        if salt:
            data["salt"] = float(salt.group("salt"))

        synergy = synergy_parser.search(data.get("label", ""))
        if synergy:
            data["synergy"] = float(synergy.group("synergy"))

        return EdhRecCardItem.model_validate(
            {
                **data,
                "tag": tag,
                "url": f"/pages{data.get('url')}.json",
                "slug": slugify(data.get("name", "")),
            }
        )


class EdhRecCardList(MightstoneModel):
    tag: str
    items: List[EdhRecCardItem] = []

    @classmethod
    def parse_payload(cls, data: dict):
        tag = data.get("tag")
        return EdhRecCardList.model_validate(
            {
                "tag": tag,
                "items": list(
                    EdhRecCardItem.parse_payload(item, tag).dict()
                    for item in data["cardviews"]
                ),
            }
        )


class EdhRecCommander(MightstoneModel):
    card: EdhRecCard
    articles: List[dict] = []
    cards: List[EdhRecCardList] = []
    mana_curve: Dict[int, int] = {i: 0 for i in range(0, 11)}
    themes: List[EdhRecCommanderSub] = []
    budget: List[EdhRecCommanderSub] = []
    distribution: EdhRecCommanderDistribution
    links: List[dict] = []

    @classmethod
    def parse_payload(cls, data: dict):
        return EdhRecCommander(
            card=EdhRecCard.model_validate(
                data.get("container", {}).get("json_dict", {}).get("card")
            ),
            cards=[
                EdhRecCardList.parse_payload(payload)
                for payload in data.get("container", {})
                .get("json_dict", {})
                .get("cardlists")
            ],
            articles=data.get("panels", {}).get("articles", []),
            links=data.get("panels", {}).get("links", []),
            mana_curve=data.get("panels", {}).get("mana_curve", {}),
            themes=data.get("panels", {}).get("tribelinks", {}).get("themes", {}),
            budget=data.get("panels", {}).get("tribelinks", {}).get("budget", {}),
            distribution=data,  # type: ignore
        )


class EdhRecFilterOperator(Enum):
    GREATER_THAN = "gt"
    LESS_THAN = "lt"
    EQUAL = "eq"
    NOT_EQUAL = "ne"


class EdhRecFilterType(Enum):
    CREATURE = "c"
    INSTANT = "i"
    SORCERY = "s"
    ARTIFACT = "a"
    ENCHANTMENT = "e"
    PLANESWALKER = "p"
    LANDS = "l"
    PRICE = "d"


class EdhRecFilterComparator(MightstoneModel):
    value: int = 0
    operator: EdhRecFilterOperator = EdhRecFilterOperator.EQUAL

    def __str__(self):
        return f"{self.operator.value}={self.value}"


class EdhRecFilterQuery(MightstoneModel):
    card_in: List[str] = []
    card_out: List[str] = []
    count: Dict[EdhRecFilterType, EdhRecFilterComparator] = {}

    def __str__(self):
        filters = []
        filters.extend([f"Out={card}" for card in self.card_out])
        filters.extend([f"In={card}" for card in self.card_in])
        filters.extend(
            [f"{field.value}:{comparator}" for field, comparator in self.count.items()]
        )
        return ";".join(filters)


class EdhRecRecs(MightstoneModel):
    commanders: List[EdhRecCard] = []
    inRecs: List[EdhRecCard] = []
    outRecs: List[EdhRecCard] = []


def slugify(string: Optional[str]):
    import slugify

    if string is None:
        return None
    return slugify.slugify(
        string, separator="-", replacements=[("'", ""), ("+", "plus-")]
    )



import asyncio
import logging
import os
import pathlib

import ijson as ijson_module
import mongomock_motor
import motor.motor_asyncio
from appdirs import AppDirs
from httpx_cache import AsyncCacheControlTransport, BaseCache, DictCache, FileCache
from injector import Binder, Module, SingletonScope, provider, singleton

from .config import DbImplem, InMemorySettings, MightstoneSettings
from .injector import cleaned
from .services.cardconjurer import CardConjurer
from .services.edhrec import EdhRecApi, EdhRecStatic
from .services.mtgjson import MtgJson
from .services.scryfall import Scryfall
from .services.wiki.api import Wiki
from .services.wotc import RuleExplorer
from .storage import MightstoneInMemoryContext, Mongod
from .types import MightstoneIjsonBackend

logger = logging.getLogger("mightstone")


class JSON(Module):
    @provider
    @singleton
    def ijson(self, config: MightstoneSettings) -> MightstoneIjsonBackend:
        try:
            out = ijson_module.get_backend(config.ijson)
            return out
        except ImportError:
            return ijson_module.get_backend("python")


class Storage(Module):
    def configure(self, binder: Binder):
        binder.bind(Mongod, scope=cleaned, to=self.mongod)
        binder.bind(motor.motor_asyncio.AsyncIOMotorClient, to=self.client)
        binder.bind(motor.motor_asyncio.AsyncIOMotorDatabase, to=self.database)
        binder.bind(mongomock_motor.AsyncMongoMockClient, to=self.mock_client)
        binder.bind(mongomock_motor.AsyncMongoMockDatabase, to=self.mock_database)

    @provider
    @singleton
    def client(
        self, config: MightstoneSettings, mongod: Mongod
    ) -> motor.motor_asyncio.AsyncIOMotorClient:
        if config.storage.implementation == DbImplem.FAKE:
            raise NotImplementedError("No real client available in fake mode")
        if config.storage.implementation == DbImplem.MOTOR:
            return motor.motor_asyncio.AsyncIOMotorClient(str(config.storage.uri))
        return motor.motor_asyncio.AsyncIOMotorClient(mongod.connection_string)

    @provider
    @singleton
    def mock_client(
        self, config: MightstoneSettings
    ) -> mongomock_motor.AsyncMongoMockClient:
        if config.storage.implementation == DbImplem.FAKE:
            return mongomock_motor.AsyncMongoMockClient(
                "mongodb://example.com:27677/tutu"
            )
        raise NotImplementedError("No fake client available in real mode")

    @provider
    def database(
        self,
        config: MightstoneSettings,
        client: motor.motor_asyncio.AsyncIOMotorClient,
        mongod: Mongod,
    ) -> motor.motor_asyncio.AsyncIOMotorDatabase:
        if config.storage.implementation == DbImplem.FAKE:
            raise NotImplementedError("No fake database available in real mode")

        # Ensure that the DB is started in case of local db
        if mongod:
            mongod.start()

        return client.get_database(config.storage.database)

    @provider
    def mock_database(
        self,
        config: MightstoneSettings,
        client: mongomock_motor.AsyncMongoMockClient,
    ) -> mongomock_motor.AsyncMongoMockClient:
        if config.storage.implementation == DbImplem.FAKE:
            return client.get_database(config.storage.database)
        raise NotImplementedError("No real database available in fake mode")

    @provider
    @singleton
    def mongod(self, appdirs: AppDirs, config: MightstoneSettings) -> Mongod:
        if not isinstance(config.storage, InMemorySettings):
            return None  # type: ignore

        directory = appdirs.user_data_dir
        if config.storage.directory:
            directory = config.storage.directory
        context = MightstoneInMemoryContext(
            directory, appdirs.user_cache_dir, str(config.storage.database)
        )

        if context.mongod_data_folder:
            config.storage.directory = pathlib.Path(context.mongod_data_folder)

        return Mongod(context)


class Httpx(Module):
    @provider
    @singleton
    def cache_backend(self, config: MightstoneSettings, appdirs: AppDirs) -> BaseCache:
        if not config.http.cache.persist:
            return DictCache()

        if not config.http.cache.directory:
            logger.debug(
                "http cache directory is not defined, using mightstone cache directory"
            )
            config.http.cache.directory = pathlib.Path(appdirs.user_cache_dir).joinpath(
                "http"
            )

        if not config.http.cache.directory.exists():
            logger.warning(
                "http cache directory %s does not exist yet, attempting to create it",
                config.http.cache.directory,
            )
            os.makedirs(config.http.cache.directory)

        return FileCache(cache_dir=config.http.cache.directory)

    @provider
    @singleton
    def cache_transport(
        self, config: MightstoneSettings, cache_backend: BaseCache
    ) -> AsyncCacheControlTransport:
        return AsyncCacheControlTransport(
            cache=cache_backend,
            cacheable_methods=tuple(config.http.cache.methods),
            cacheable_status_codes=tuple(config.http.cache.status),
        )


class Services(Module):
    @singleton
    @provider
    def rule_explorer(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> RuleExplorer:
        return RuleExplorer(transport=cache, ijson=ijson)  # type: ignore

    @singleton
    @provider
    def scryfall(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> Scryfall:
        return Scryfall(transport=cache, ijson=ijson)  # type: ignore

    @singleton
    @provider
    def edhrec_static(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> EdhRecStatic:
        return EdhRecStatic(transport=cache, ijson=ijson)  # type: ignore

    @singleton
    @provider
    def edhrec_api(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> EdhRecApi:
        return EdhRecApi(transport=cache, ijson=ijson)  # type: ignore

    @singleton
    @provider
    def card_conjurer(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> CardConjurer:
        return CardConjurer(transport=cache, ijson=ijson)

    @singleton
    @provider
    def mtg_json(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> MtgJson:
        return MtgJson(transport=cache, ijson=ijson)

    @singleton
    @provider
    def wiki(
        self,
        cache: AsyncCacheControlTransport,
        ijson: MightstoneIjsonBackend,
    ) -> Wiki:
        return Wiki(transport=cache, ijson=ijson)


class Configuration(Module):
    def configure(self, binder: Binder):
        binder.bind(
            MightstoneSettings, scope=SingletonScope, to=lambda: MightstoneSettings()
        )


class Directories(Module):
    @provider
    @singleton
    def app_dirs(self, config: MightstoneSettings) -> AppDirs:
        return AppDirs(config.appname)


modules = [JSON, Configuration, Directories, Storage, Httpx]



import asyncio
from typing import Optional

import ijson as ijson_module
from httpx import AsyncClient
from httpx_cache import AsyncCacheControlTransport
from injector import inject

from ..types import MightstoneIjsonBackend


class ServiceError(Exception):
    def __init__(self, message, url=None, status=None, data=None, method=None):
        self.message = message
        self.url = url
        self.status = status
        self.data = data
        self.method = method

    def __str__(self):
        return "{message} (HTTP:{status} {method} {url})".format(**self.__dict__)


class MightstoneHttpClient:
    base_url: str
    """
    Base url of the service (must be a root path such as https://example.com)
    """
    delay = 0
    """
    Induced delay in second between each API call
    """

    @inject
    def __init__(
        self,
        transport: Optional[AsyncCacheControlTransport] = None,
        ijson: Optional[MightstoneIjsonBackend] = None,
    ):
        self.transport = transport
        self.ijson = ijson or ijson_module.get_backend(ijson_module.backend)

    @property
    def client(self):
        # See: https://github.com/encode/httpx/issues/2473
        options = {
            "transport": self.transport,
            "headers": {"cache-control": f"max-age={60 * 60 * 24}"},
        }
        if hasattr(self, "base_url"):
            options["base_url"] = self.base_url

        return AsyncClient(**options)

    async def close(self):
        await self.client.aclose()

    async def _sleep(self):
        await asyncio.sleep(self.delay)



import re
from collections import defaultdict
from datetime import datetime
from io import StringIO
from itertools import takewhile
from typing import Any, DefaultDict, Dict, List, Mapping, Optional, TextIO
from uuid import UUID

from pydantic import ConfigDict, GetCoreSchemaHandler, model_validator
from pydantic_core import CoreSchema, core_schema

from mightstone.common import generate_uuid_from_string
from mightstone.core import (
    MightstoneDocument,
    MightstoneModel,
    MightstoneSerializableDocument,
)


class RuleRef(str):
    """
    Rules reference use the same pattern with a rule number

    <rule>[.<sub_rule>[<letter>]][<trailing_dot>]

     * rule: an integer of three digit, first digit matches parent section
     * sub_rule: an integer
     * letter: a letter (o, and l are not valid)
     * trailing dot: used in table of content for readability purpose
       and then used inconsistently in the rule themselves.

    RuleRef consider the trailing dot a legal, but won’t keep it in its canonical
    notation
    """

    regex = re.compile(
        r"(?P<reference>"
        r"(?P<rule>\d{3})"
        r"(\.((?P<sub_rule>\d+)(?P<letter>[a-z])?))?"
        r"(?P<trailing_dot>\.)?"
        r")"
    )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, value, *args, **kwargs):
        return super(RuleRef, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.regex.match(value)
        if not res:
            raise ValueError(f"{self} is not a valid reference")
        if res.group("letter") in ["l", "o"]:
            raise ValueError(
                f"{self} is not a valid reference, letters (o, l) are invalid"
            )

        self.rule = int(res.group("rule"))
        self.sub_rule = None
        if res.group("sub_rule"):
            self.sub_rule = int(res.group("sub_rule"))
        self.letter = res.group("letter")
        self.canonical = self.build(self.rule, self.sub_rule, self.letter)
        self.section = int(res.group("rule")[0])

    @classmethod
    def build(
        cls, rule: int, sub_rule: Optional[int] = None, letter: Optional[str] = None
    ):
        out = str(rule)
        if sub_rule:
            out += f".{sub_rule}"
            if letter:
                out += f"{letter}"
        return out

    def next(self):
        if self.letter and self.sub_rule and self.rule:
            increment = 1
            if self.letter in ["k", "n"]:
                increment = 2
            return RuleRef(
                self.build(self.rule, self.sub_rule, chr(ord(self.letter) + increment))
            )

        if self.sub_rule and self.rule:
            return RuleRef(self.build(self.rule, self.sub_rule + 1))

        return RuleRef(self.build(self.rule + 1))

    def prev(self):
        if self.letter == "a":
            return None
        if self.sub_rule == 1 and not self.letter:
            return None
        if self.rule == 100 and not self.sub_rule:
            return None

        if self.letter and self.sub_rule and self.rule:
            increment = 1
            if self.letter in ["m", "p"]:
                increment = 2

            return RuleRef(
                self.build(self.rule, self.sub_rule, chr(ord(self.letter) - increment))
            )

        if self.sub_rule and self.rule:
            return RuleRef(self.build(self.rule, self.sub_rule - 1))

        return RuleRef(self.build(self.rule - 1))

    def __eq__(self, other):
        try:
            return self.canonical == other.canonical
        except AttributeError:
            return self.canonical == other

    def __le__(self, other):
        return self.canonical <= other.canonical


class SectionRef(str):
    regex = re.compile(r"(?P<reference>(?P<section>\d)\.?)")

    def __new__(cls, value, *args, **kwargs):
        return super(SectionRef, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.regex.match(value)
        if not res:
            raise ValueError(f"{self} is not a valid reference")
        self.section = int(res.group("section"))


class RuleText(str):
    """
    A string than can contain reference to rule
    """

    see_rule = re.compile(r"rule " + RuleRef.regex.pattern)
    see_section = re.compile(r"section " + SectionRef.regex.pattern)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, value, *args, **kwargs):
        return super(RuleText, cls).__new__(cls, value)

    def __init__(self, value):
        self.refs = []
        for item in self.see_rule.findall(value):
            self.refs.append(RuleRef(item[0]))
        for item in self.see_section.findall(value):
            self.refs.append(SectionRef(item[0]))


class Example(str):
    pattern = re.compile(r"(?P<example>Example: (?P<text>.+))")

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, value, *args, **kwargs):
        return super(Example, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.pattern.match(value)
        if not res:
            raise ValueError("String is not an example")
        self.text = RuleText(res.group("text"))


class Rule(MightstoneModel):
    ref: RuleRef
    text: RuleText
    examples: List[Example] = []

    @classmethod
    def parse_text(cls, value):
        pattern = re.compile(RuleRef.regex.pattern + r"\s+(?P<text>\w+.*)")

        res = pattern.match(value)
        if not res:
            raise ValueError("String is not a rule")

        return Rule(
            ref=RuleRef(res.group("reference")), text=RuleText(res.group("text"))
        )


class Effectiveness(str):
    pattern = re.compile(
        r"(?P<effective>These rules are effective as of"
        r" (?P<date>(?P<month>\w+) (?P<day>\d+), (?P<year>\d{4})).)"
    )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __new__(cls, value, *args, **kwargs):
        return super(Effectiveness, cls).__new__(cls, value)

    def __init__(self, value):
        res = self.pattern.match(value)
        if not res:
            raise ValueError("String is not a valid effectiveness string")
        self.date = datetime.strptime(res.group("date"), "%B %d, %Y").date()


class Ruleset(MightstoneModel):
    rules: Dict[str, Rule] = dict()
    last_rule: Optional[Rule] = None

    # def __iter__(self) -> Iterator[Rule]:  # type: ignore
    #     for rule in self.rules.values():
    #         yield rule

    def __getitem__(self, k: str) -> Rule:
        return self.rules[k]

    def __len__(self) -> int:
        return len(self.rules)

    def parse_text(self, value: str):
        for line in value.splitlines():
            try:
                rule = Rule.parse_text(line)
                self.rules[rule.ref.canonical] = rule
                self.last_rule = rule
                continue
            except ValueError:
                pass

            try:
                if self.last_rule:
                    self.rules[self.last_rule.ref.canonical].examples.append(
                        Example(line)
                    )
                continue
            except (ValueError, AttributeError):
                pass

    def search(self, string: str):
        return [
            item for item in self.rules.values() if string.lower() in item.text.lower()
        ]

    def range(self, low: str, up: Optional[str] = None):
        low = RuleRef(low)
        if not up:
            up = RuleRef(low).next()
        else:
            up = RuleRef(up)

        if up:
            return [item for item in self.rules.values() if up > item.ref >= low]

        return [item for item in self.rules.values() if item.ref >= low]

    def index(self):
        self.rules = dict(sorted(self.rules.items()))


class Term(MightstoneModel):
    term: str
    description: RuleText


class Glossary(MightstoneModel, Mapping):
    terms: Dict[str, Term] = {}

    def __getitem__(self, k: str) -> Term:
        return self.terms[k.lower()]

    def __len__(self) -> int:
        return len(self.terms)

    def add(self, term, text):
        self.terms[term.lower()] = Term(description=RuleText(text), term=term)

    def search(self, string):
        return [
            item
            for item in self.terms.values()
            if string.lower() in item.term.lower()
            or string.lower() in item.description.lower()
        ]

    def index(self):
        self.terms = dict(sorted(self.terms.items()))


class ComprehensiveRules(MightstoneDocument):
    id: Optional[UUID] = None  # type: ignore
    effective: Optional[Effectiveness] = None
    ruleset: Ruleset = Ruleset()
    glossary: Glossary = Glossary()

    @model_validator(mode="wrap")
    @classmethod
    def enforce_id(cls, value: Any, handler) -> "ComprehensiveRules":
        doc = handler(value)

        if not isinstance(value, Dict):
            return doc

        if doc.id:
            return doc

        if "id" not in value or not value["id"]:
            if "effective" in value:
                doc.id = generate_uuid_from_string(value["effective"])

        return doc

    def search(self, string):
        found = []
        found.extend(self.ruleset.search(string))
        found.extend(self.glossary.search(string))
        return found

    @classmethod
    def parse(cls, buffer: TextIO):
        cr = ComprehensiveRules()
        in_glossary = False
        in_credits = False
        buffer2 = StringIO("\n".join(buffer.read().splitlines()))

        for line in buffer2:
            line = line.strip()
            if not line:
                # Ignore empty lines
                continue

            if not cr.effective:
                # No need to search for effectiveness once found
                try:
                    cr.effective = Effectiveness(line)
                    cr.id = generate_uuid_from_string(cr.effective)
                    continue
                except ValueError:
                    ...

            if not in_glossary:
                if "100.1" in cr.ruleset.rules and line == "Glossary":
                    in_glossary = True
                    continue

                cr.ruleset.parse_text(line)
                continue

            if not in_credits:
                if len(cr.glossary.terms) and line == "Credits":
                    in_credits = True
                    continue

                text = "\n".join(
                    [x.strip() for x in takewhile(lambda x: x.strip() != "", buffer2)]
                )
                cr.glossary.add(line, text)
                continue

        cr.ruleset.index()
        cr.glossary.index()

        return cr

    def diff(self, cr: "ComprehensiveRules"):
        """
        Compare two comprehensive rule set for change in rules and terms

        For both, terms and rules, provide a dict for:
         - `added` a dict of added object
         - `changed` a dict of dict (before/after) object
         - `removed` a dict of removed object

        :param cr: An other comprehensive rule set to compare
        :return: a dict of changes
        """
        if cr.effective and self.effective and cr.effective > self.effective:
            older = self
            newer = cr
        else:
            older = cr
            newer = self

        diff: DefaultDict = defaultdict(
            lambda: {"added": {}, "removed": {}, "changed": {}}
        )

        new_rules = set(newer.ruleset.rules)
        old_rules = set(older.ruleset.rules)
        moved_from = {}

        # Check for inclusion, search for same text, but new index
        for ref in new_rules - old_rules:
            current = RuleRef(ref)
            previous = current.prev()
            while previous:
                if (
                    older.ruleset[previous.canonical].text
                    == newer.ruleset[current.canonical].text
                ):
                    moved_from[previous.canonical] = current.canonical
                    diff["rules"]["changed"][current.canonical] = {
                        "before": older.ruleset[previous.canonical],
                        "after": newer.ruleset[current.canonical],
                    }
                    current = previous
                    previous = current.prev()
                else:
                    diff["rules"]["added"][current.canonical] = newer.ruleset[
                        current.canonical
                    ]
                    break
        for ref in old_rules - new_rules:
            diff["rules"]["removed"][ref] = older.ruleset[ref]
        for ref in new_rules.intersection(old_rules):
            if newer.ruleset[ref].text != older.ruleset[ref].text:
                if ref not in moved_from:
                    diff["rules"]["changed"][ref] = {
                        "before": older.ruleset[ref],
                        "after": newer.ruleset[ref],
                    }

        new_terms = set(newer.glossary.terms)
        old_terms = set(older.glossary.terms)
        for ref in new_terms - old_terms:
            diff["terms"]["added"][ref] = newer.glossary[ref]
        for ref in old_terms - new_terms:
            diff["terms"]["removed"][ref] = older.glossary[ref]
        for ref in new_terms.intersection(old_terms):
            if newer.glossary[ref].description != older.glossary[ref].description:
                diff["terms"]["changed"][ref] = {
                    "before": older.glossary[ref],
                    "after": newer.glossary[ref],
                }

        return diff


class SerializableComprehensiveRules(
    ComprehensiveRules, MightstoneSerializableDocument
):
    id: Optional[UUID] = None  # type: ignore



import typing
from enum import Enum
from typing import (
    Annotated,
    Any,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Pattern,
    Type,
    TypedDict,
    Union,
    overload,
)

from pydantic.fields import Field
from pydantic_extra_types.color import Color
from typing_extensions import Unpack

from mightstone.core import (
    MightstoneDocument,
    MightstoneModel,
    MightstoneSerializableDocument,
)


class LayerTypes(str, Enum):
    GROUP = "group"
    IMAGE = "image"
    TEXT = "text"


class Filters(str, Enum):
    COLOR_OVERLAY = "colorOverlay"
    SHADOW = "shadow"


class HorizontalAlign(str, Enum):
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"


class VerticalAlign(str, Enum):
    TOP = "top"
    CENTER = "center"
    BOTTOM = "bottom"


class BoundType(str, Enum):
    FIT = "fit"
    FILL = "fill"


class Tags(Enum):
    """
    Another important note is that many text and image layers have the tags "editable",
    and there are two groups tagged "new-image-group" and "new-text-group". These tags
    tell the UI which layers should be presented as editable, and which layer groups
    should store newly-added layers.
    """

    EDITABLE = "editable"
    DRAGGABLE = "draggable"
    NEW_IMAGE_GROUP = "new-image-group"
    NEW_TEXT_GROUP = "new-text-group"
    NAME = "name"


class TemplateRef(MightstoneModel):
    url: str


class Dependencies(MightstoneModel):
    extensions: List[str] = []
    template: Optional[TemplateRef] = None


class Bound(MightstoneModel):
    x: int
    y: int
    width: int
    height: int
    type: BoundType
    horizontal: HorizontalAlign
    vertical: VerticalAlign


class LayerSearch(TypedDict):
    model: Optional[Type[MightstoneModel]]
    tag: Optional[Tags]
    name: Optional[Union[str, Pattern]]
    type: Optional[LayerTypes]


class ImageSearch(TypedDict):
    model: Optional[Type[MightstoneModel]]
    tag: Optional[Tags]
    name: Optional[Union[str, Pattern]]
    type: Literal[LayerTypes.IMAGE]


class TextSearch(TypedDict):
    model: Optional[Type[MightstoneModel]]
    tag: Optional[Tags]
    name: Optional[Union[str, Pattern]]
    type: Literal[LayerTypes.TEXT]


class GroupSearch(TypedDict):
    model: Optional[Type[MightstoneModel]]
    tag: Optional[Tags]
    name: Optional[Union[str, Pattern]]
    type: Literal[LayerTypes.GROUP]


class Layer(MightstoneModel):
    type: Any
    name: Optional[str] = None
    tags: Optional[List[Tags]] = None

    @overload
    def find_all(
        self, **kwargs: Unpack[ImageSearch]
    ) -> Generator[Union["Image", "Mask"], None, None]: ...

    @overload
    def find_all(
        self, **kwargs: Unpack[TextSearch]
    ) -> Generator["Text", None, None]: ...
    @overload
    def find_all(
        self, **kwargs: Unpack[GroupSearch]
    ) -> Generator["Group", None, None]: ...

    @overload
    def find_all(
        self, **kwargs: Unpack[LayerSearch]
    ) -> Generator[Union["Image", "Group", "Text", "Mask"], None, None]: ...

    def find_all(
        self, **kwargs: Unpack[LayerSearch]
    ) -> Generator[Union["Image", "Group", "Text", "Mask"], None, None]:
        # TODO: sort by Z index by default
        for layer in self._recurse():
            model = kwargs.get("model")
            if model:
                if not isinstance(layer, model):
                    continue

            name = kwargs.get("name")
            if isinstance(name, Pattern):
                try:
                    if not layer.name:
                        continue
                    elif not name.match(layer.name):
                        continue
                except TypeError:
                    continue
            elif name and layer.name != name:
                continue

            tag = kwargs.get("tag")
            if tag:
                if not hasattr(layer, "tags"):
                    continue

                if not layer.tags:
                    continue

                if tag not in layer.tags:
                    continue

            typ = kwargs.get("type")
            if typ and typ != layer.type:
                continue
            yield layer

    @overload
    def find_one(
        self, **kwargs: Unpack[ImageSearch]
    ) -> Optional[Union["Image", "Mask"]]: ...
    @overload
    def find_one(self, **kwargs: Unpack[TextSearch]) -> Optional["Text"]: ...
    @overload
    def find_one(self, **kwargs: Unpack[GroupSearch]) -> Optional["Group"]: ...

    def find_one(
        self, **kwargs: Unpack[LayerSearch]
    ) -> Optional[Union["Image", "Group", "Text", "Mask"]]:
        it = self.find_all(**kwargs)
        return next(it, None)

    @typing.no_type_check
    def _recurse(self) -> Generator["Layer", None, None]:
        try:
            for child in self.children:
                yield from child._recurse()
        except AttributeError:
            ...
        try:
            for mask in self.masks:
                yield mask
        except (AttributeError, TypeError):
            ...

        yield self


class Mask(MightstoneModel):
    type: Literal[LayerTypes.IMAGE]
    name: str
    src: str


class Variant(MightstoneModel):
    name: str
    src: str
    thumb: str
    x: Optional[int] = None


class FilterOverlay(MightstoneModel):
    type: Literal[Filters.COLOR_OVERLAY]
    color: Color


class FilterShadow(MightstoneModel):
    type: Literal[Filters.SHADOW]
    color: Color = Color("black")
    x: int
    y: int


Filter = Annotated[Union[FilterShadow, FilterOverlay], Field(discriminator="type")]


class Image(Layer):
    """
    A layer composed of an image (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.IMAGE]
    src: str
    x: int = 0
    y: int = 0
    z: int = 0
    width: int = 100
    height: int = 100
    thumb: Optional[str] = None
    bounds: Optional[Bound] = None
    masks: Optional[List[Mask]] = None
    opacity: Optional[float] = None
    filters: Optional[List[Filter]] = None


class Text(Layer):
    """
    A layer composed of a text (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.TEXT]
    text: str
    font: Optional[str] = None
    color: Color = Color("#FFFFFF")
    oneLine: bool = False
    align: Optional[HorizontalAlign] = HorizontalAlign.LEFT
    verticalAlign: Optional[VerticalAlign] = VerticalAlign.CENTER
    lineHeightScale: float = 1
    fontWeight: Optional[str] = None  # bold
    rotation: int = 0
    size: int
    opacity: Optional[float] = None
    x: int
    y: int
    width: int
    height: int
    textCodesOnly: bool = False
    filters: Optional[List[Union[FilterShadow, FilterOverlay]]] = None


class Group(Layer):
    """
    A layer composed of a list of other layers (One of the 3 possible layer types)
    """

    type: Literal[LayerTypes.GROUP] = LayerTypes.GROUP.value  # type: ignore
    children: List["AnyLayer"] = []


AnyLayer = Annotated[Union[Group, Text, Image], Field(discriminator="type")]


class Card(MightstoneDocument):
    """
    A Card as described in Card Conjurer JSON
    """

    asset_root_url: str = ""
    """
    Not part of CardConjurer model
    This allow to re-contextualize relative path and build proper urls
    """

    name: str
    width: int
    height: int
    corners: int = 0
    marginX: int = 0
    marginY: int = 0
    dependencies: Dependencies = Dependencies()
    data: Group = Group(type=LayerTypes.GROUP)  # type: ignore

    def find_one_layer(self, **kwargs):
        return self.data.find_one(**kwargs)

    def find_many_layer(self, **kwargs):
        return self.data.find_all(**kwargs)

    def find_many_text(self, **kwargs) -> Generator[Text, None, None]:
        return self.data.find_all(type=LayerTypes.TEXT, **kwargs)

    def find_many_images(self, **kwargs) -> Generator[Union[Image, Mask], None, None]:
        return self.data.find_all(type=LayerTypes.IMAGE, **kwargs)

    def find_many_groups(self, **kwargs) -> Generator[Group, None, None]:
        return self.data.find_all(type=LayerTypes.GROUP, **kwargs)


class SerializableCard(Card, MightstoneSerializableDocument): ...


class TemplateMetaData(MightstoneModel):
    """
    This metadata is entirely optional, but it helps describe what the template is for,
    who made it, and how updated it is.
    """

    name: Optional[str] = None
    game: Optional[str] = None
    creator: Optional[str] = None
    created: Optional[str] = None  # TODO: date January 7, 2022
    updated: Optional[str] = None  # TODO: date January 7, 2022


class TemplateContextImageSet(MightstoneModel):
    """
    The goal of an image set is to describe how to add an image layer for each
    available image in your template. They can organize your images into distinct
    groups that share common traits, and can remove certain duplicated information.
    """

    prototype: Dict[str, Any]
    variants: List[Variant]
    masks: Optional[List[Mask]] = None


class TemplateFont(MightstoneModel):
    """
    Fonts are nice and simple. All you define here are the names and URLs of your
    fonts. You can also declare the "weight" and "style" if you have variants of
    fonts with the same name.
    """

    name: str
    src: str
    weight: Optional[str] = None
    style: Optional[str] = None


class Symbol(MightstoneModel, extra="allow"):
    src: str
    name: str
    scale: float = 1
    spacing: float = 0.05
    verticalShift: float = 0.05


class TemplateSymbol(MightstoneModel):
    src: str
    name: Union[List[str], str]

    def to_dict(self, prefix: str, prototype: dict, unique=False) -> Dict[str, Symbol]:
        names = [self.name] if isinstance(self.name, str) else self.name
        symbol = Symbol(src=prefix + self.src, name=names[0], **prototype)
        if unique:
            return {names[0]: symbol}
        return {name: symbol for name in names}


class SymbolSet(MightstoneModel):
    name: str
    symbols: List[TemplateSymbol]
    prototype: Dict[str, Any]
    srcPrefix: str

    def to_dict(self, unique=False) -> Dict[str, Symbol]:
        return dict(
            pair
            for s in self.symbols
            for pair in s.to_dict(self.srcPrefix, self.prototype, unique).items()
        )


class TemplateContext(MightstoneModel):
    """
    The Template Context is pretty large, but here are the main sections. We'll break
    each one down individually, in detail.
    """

    license: Optional[dict] = None
    image_sets: List[TemplateContextImageSet] = Field(alias="imageSets", default=[])
    fonts: List[TemplateFont] = []
    symbol_sets: List[SymbolSet] = Field(alias="symbolSets", default=[])

    def symbols(self, unique=False) -> Dict[str, Symbol]:
        return dict(
            (k.lower(), v)
            for ss in self.symbol_sets
            for k, v in ss.to_dict(unique).items()
        )


class Template(MightstoneDocument):
    """
    A Template as described in Card Conjurer JSON
    """

    asset_root_url: str = ""
    """
    Not part of CardConjurer model
    This allow to re-contextualize relative path and build proper urls
    """

    name: str
    context: TemplateContext
    card: Card

    @classmethod
    def dummy(cls):
        """
        :return: A dummy template, with nothing in it
        """
        return Template(
            name="Dummy",
            context=TemplateContext(),
            card=Card(name="Dummy", width=100, height=100),
        )


class SerializableTemplate(Template, MightstoneSerializableDocument): ...


Group.model_rebuild()



import itertools
import logging
from collections import Counter
from typing import (
    Annotated,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    overload,
)

from ordered_set import OrderedSet
from pydantic import GetCoreSchemaHandler, StringConstraints
from pydantic_core import CoreSchema, core_schema

from mightstone.core import MightstoneModel

logger = logging.getLogger(__name__)

ColorGlyph = Annotated[
    str, StringConstraints(min_length=1, max_length=1, to_lower=True)
]


class Color(MightstoneModel):
    symbol: ColorGlyph
    index: int

    def __str__(self):
        return f"{{{self.symbol.upper()}}}"

    def __repr__(self):
        return f"Color({self.symbol})"


class ColorPie(Sequence[Color]):
    def __init__(self, colors: Iterable[Color]):
        if not all(isinstance(x, Color) for x in colors):
            raise ValueError("Please provide a Color object iterable")

        duplicates = [k for k, v in Counter(list(colors)).items() if v > 1]
        if len(duplicates):
            duplicates_as_string = ",".join(map(str, duplicates))
            raise ValueError(
                f"A color pie cannot hold the same color twice. {duplicates_as_string}"
            )

        # TODO: search duplicates color values
        self.colors = list(colors)

    def __getitem__(self, i: Union[int, slice, Union[int, str]]):
        if isinstance(i, int):
            return self.colors[i]

        try:
            return next(color for color in self.colors if color.symbol == i)
        except StopIteration:
            raise KeyError(f"{i} not found in {self}")

    def __len__(self) -> int:
        return len(self.colors)

    def __iter__(self) -> Iterator[Color]:
        for color in self.colors:
            yield color

    def shift(self, color: Color, step: int = 1) -> Color:
        return self.colors[(step + self.index(color)) % len(self.colors)]

    def __hash__(self):
        return hash(tuple(self))

    def parse(self, identity_as_string: str) -> "Identity":
        colors = []
        for letter in identity_as_string:
            colors.append(self[letter])
        return Identity(colors)

    def combinations(self) -> List["Identity"]:
        """
        A mathematical computation of all possible combinations of colors
        This will not provide a proper color pie centric combination though
        and cannot be used to provide a complete identity map that would
        build the red enemy guild (Boros) as rw, instead of wr in this case
        """
        return [
            Identity(c)
            for length in range(0, len(self.colors) + 1)
            for c in itertools.combinations(self.colors, length)
        ]

    def build_identity_map(self) -> "IdentityMap":
        idmap = IdentityMap(self)

        for combination in self.combinations():
            idmap.add(combination)

        return idmap

    def __repr__(self):
        return f"ColorPie({self.colors})"


class Identity(Sequence[Color]):
    def __init__(self, colors: Iterable[Color]):
        self.colors = OrderedSet(colors)
        self._name = ""
        self.aliases: List[str] = []
        self._canonical: Optional[str] = None

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def describe(
        self,
        name: Optional[str] = None,
        aliases: Optional[List[str]] = None,
        canonical: Optional[str] = None,
    ):
        if name:
            self._name = name
        if aliases:
            self.aliases.extend(aliases)
        if canonical:
            self._canonical = canonical

    @property
    def name(self):
        if not self._name:
            return self.canonical
        return self._name

    @property
    def canonical(self) -> str:
        if self._canonical:
            return self._canonical
        if len(self.colors) == 0:
            return ""
        return "".join(
            [color.symbol for color in sorted(self.colors, key=lambda x: x.index)]
        )

    def checksum(self) -> int:
        """Checksum is computed from binary position of the color in the color-pie"""
        return sum(1 << color.index for color in self.colors)

    def matches(self, k: str):
        search = k.lower()
        if search in self.name.lower():
            return True
        if search in map(lambda x: x.lower(), self.aliases):
            return True
        return False

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Identity({self.canonical})"

    @overload
    def __getitem__(self, i: int) -> Color: ...

    @overload
    def __getitem__(self, i: slice) -> OrderedSet[Color]: ...

    def __getitem__(self, i: Union[int, slice]) -> Union[Color, OrderedSet[Color]]:
        return self.colors.__getitem__(i)

    def __len__(self):
        return len(self.colors)

    def __eq__(self, other: object):
        if not isinstance(other, Identity):
            return NotImplemented

        return other.checksum() == self.checksum()


class IdentityMap(Mapping[int, Identity]):
    def __init__(self, pie: ColorPie):
        self.map: Dict[int, Identity] = {}
        self.pie = pie

    def refine(
        self,
        canonical: str,
        name: Optional[str] = None,
        aliases: Optional[List[str]] = None,
    ):
        self[canonical].describe(name=name, aliases=aliases, canonical=canonical)

    def add(self, colors=Iterable[Color]):
        """
        Appends an identity to the map
        No addition if the identity already exists
        """
        ident = Identity(colors)
        if ident.checksum() not in self.map:
            self.map[ident.checksum()] = ident

    def __getitem__(self, k: Union[int, str, Identity]) -> Identity:
        if isinstance(k, int):
            return self.map[k]

        if isinstance(k, Identity):
            return self.map[k.checksum()]

        if not isinstance(k, str):
            raise KeyError

        try:
            match = self.pie.parse(k)
            return self.map[match.checksum()]
        except KeyError:
            for identity in self.map.values():
                if identity.matches(k):
                    return identity

        raise KeyError

    def __len__(self) -> int:
        return self.map.__len__()

    def __iter__(self) -> Iterator[int]:
        return self.map.__iter__()


class ColorAffinity(MightstoneModel):
    """
    When talking about which colors get which evergreen creature keywords,
    R&D tends to talk about a system called "primary/secondary/tertiary".
    In their quest to differentiate the colors in the color wheel,
    each should have strengths and weaknesses.
    """

    primary: List[Color] = []
    secondary: List[Color] = []
    tertiary: List[Color] = []



import importlib
from pathlib import Path
from typing import Optional

import click

from mightstone.cli.models import MightstoneCli, pass_mightstone


@click.group()
def wotc():
    pass


@wotc.group()
def scrape(): ...


module = importlib.import_module("mightstone.rule.data")
default_output = Path(module.__path__[0]).joinpath("rules.json")


@scrape.command()
@click.option("--url", "-U", type=str, default=None)
@click.argument("output", type=click.File("w"), default=default_output)
@pass_mightstone
def rules(mightstone: MightstoneCli, url: Optional[str], output: click.utils.LazyFile):
    print("Building rules from mtg website...")
    rules = mightstone.app.rule_explorer.open(url)

    print("Saving informations...")
    output.write(rules.model_dump_json(indent=2))
    print(f"Rules saved into {output.name}")



import json
import logging
import os
import pathlib
from enum import Enum
from typing import Literal, Optional, Tuple, Type, Union

import toml
import yaml
from pydantic.networks import MongoDsn
from pydantic_settings import (
    BaseSettings,
    InitSettingsSource,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
    YamlConfigSettingsSource,
)
from slugify import slugify

from mightstone.core import MightstoneError

logger = logging.getLogger("mightstone")


class DbImplem(str, Enum):
    LOCAL = "local"
    MOTOR = "motor"
    FAKE = "fake"


class IjsonEnum(str, Enum):
    YAJL2_C = "yajl2_c"
    YAJL2_CFFI = "yajl2_cffi"
    YAJL2 = "yajl2"
    YAJL = "yajl"
    PYTHON = "python"


class HttpCacheSettings(BaseSettings):
    persist: bool = True
    directory: Optional[pathlib.Path] = None
    methods: list[str] = ["GET"]
    status: list[int] = [200, 203, 300, 301, 308]


class HttpSettings(BaseSettings):
    cache: HttpCacheSettings = HttpCacheSettings()


class InMemorySettings(BaseSettings):
    implementation: Literal[DbImplem.LOCAL] = DbImplem.LOCAL
    directory: Optional[pathlib.Path] = None
    database: str = "mightstone"


class MotorSettings(BaseSettings):
    implementation: Literal[DbImplem.MOTOR] = DbImplem.MOTOR
    uri: MongoDsn
    database: str = "mightstone"


class FakeSettings(BaseSettings):
    implementation: Literal[DbImplem.FAKE] = DbImplem.FAKE
    database: str = "mightstone"


class MightstoneSettings(BaseSettings):
    appname: str = "Mightstone"
    storage: Union[InMemorySettings, MotorSettings, FakeSettings] = InMemorySettings()
    http: HttpSettings = HttpSettings()
    ijson: IjsonEnum = IjsonEnum.PYTHON

    model_config = SettingsConfigDict(
        {
            "env_prefix": "mightstone_",
            "env_nested_delimiter": "__",
            "env_file_encoding": "utf-8",
        }
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            SettingsSourceGenerator("mightstone").build(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


class SettingsSourceGenerator:
    def __init__(
        self, app_name: str = "mightstone", env_variable: Optional[str] = None
    ):
        self.app_name = slugify(app_name, separator="_")
        if not env_variable:
            env_variable = f"{self.app_name.upper()}_CONFIG_FILE"
        self.env_variable = env_variable

    def candidate_paths(
        self,
    ) -> list[pathlib.Path]:
        candidate = os.getenv(self.env_variable)
        if candidate:
            return [pathlib.Path(candidate)]

        return [
            pathlib.Path(path).joinpath(f"{self.app_name}.{extension}")
            for path in [os.getcwd(), os.path.expanduser("~")]
            for extension in ["yaml", "yml", "json", "toml"]
        ]

    @classmethod
    def load_file(cls, filepath, encoding):
        try:
            with open(filepath, "r", encoding=encoding) as f:
                _, extension = os.path.splitext(filepath)
                if extension.lower() in [".yml", ".yaml"]:
                    return yaml.safe_load(f)

                if extension.lower() == ".toml":
                    return toml.load(f)

                return json.load(f)
        except Exception as e:
            raise MightstoneError("Unable to parse %s" % filepath) from e

    def build(self, settings_cls: Type[BaseSettings]) -> Union[
        InitSettingsSource,
        JsonConfigSettingsSource,
        YamlConfigSettingsSource,
        TomlConfigSettingsSource,
    ]:
        for candidate in self.candidate_paths():
            if not candidate.exists():
                continue

            _, extension = os.path.splitext(candidate)
            try:
                if extension.lower() in [".yml", ".yaml"]:
                    return YamlConfigSettingsSource(settings_cls, candidate)

                if extension.lower() == ".toml":
                    return TomlConfigSettingsSource(settings_cls, candidate)

                return JsonConfigSettingsSource(settings_cls, candidate)
            except MightstoneError as e:
                logger.fatal("Unable to parse configuration from %s, %s", candidate, e)

        return InitSettingsSource(settings_cls, {})



"""
MTGJSON support core
"""

import json
import logging
from enum import Enum
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import asyncstdlib
from httpx import HTTPStatusError
from pydantic import ValidationError

from mightstone.ass import compressor, sync_generator, synchronize
from mightstone.core import MightstoneModel
from mightstone.services import MightstoneHttpClient, ServiceError
from mightstone.services.mtgjson.models import (
    Card,
    CardAtomic,
    CardFace,
    CardPrices,
    CardTypes,
    Deck,
    DeckList,
    Keywords,
    Meta,
    Set,
    SetList,
    TcgPlayerSKU,
    TcgPlayerSKUs,
)

DictOfListOfKey = Tuple[str, int]
DictOfListOfModel = Tuple[DictOfListOfKey, Any]
ListOfKey = int
ListOfModel = Tuple[ListOfKey, Any]
DictOfKey = str
DictOfModel = Tuple[DictOfKey, Any]
GeneratorModel = Union[DictOfListOfModel, ListOfModel, DictOfModel]
GeneratorKey = Union[DictOfListOfKey, ListOfKey, DictOfKey]

logger = logging.getLogger("mightstone")


class MtgJsonMode(Enum):
    """
    Available data parse mode

    MTGJSON model is not consistent, this enum describe the expected data structure
    of a MTGJSON response
    """

    LIST_OF_MODEL = 0
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": [{"prop": 1}, "b": {"prop": 2}]}
    """
    DICT_OF_MODEL = 1
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": {"a": {"prop": 1}, "b": {"prop": 2}}}
    """
    DICT_OF_LIST_OF_MODEL = 2
    """
    In this mode, we expect a structure similar to
     .. code-block:: json
     {"data": {"a": [{"prop": 1}], "b": [{"prop": 2}]}}
    """


class MtgJsonCompression(str, Enum):
    """
    Available compression mode enumerator

    MTGJSON provide 5 compression formats, Mightstone support 4 of them.
    """

    NONE = ""
    """ No compression, use raw JSON """
    XZ = "xz"
    """ LZMA compression, use .xz files """
    ZIP = "zip"
    """ ZIP compression, use .zip files (not supported)"""
    GZIP = "gz"
    """ GZIP compression, use .gz files"""
    BZ2 = "bz2"
    """ BZIP2 compression, use .bz2 files"""

    def to_stream_compression(self):
        """
        Compute the compression type to a python module

        :return: the name of the python module
        """
        if self.value == "":
            return None
        elif self.value == "xz":
            return "lzma"
        elif self.value == "gz":
            return "gzip"
        elif self.value == "bz2":
            return "bzip2"
        raise ValueError(f"{self.name} compression protocol cannot be read as a stream")


_T = TypeVar("_T", bound=MightstoneModel)


class MtgJson(MightstoneHttpClient):
    """
    MTGJSON client

    Supports compression and will get gzip versions by default.
    """

    base_url = "https://mtgjson.com"

    def __init__(
        self,
        compression: Optional[MtgJsonCompression] = None,
        version: int = 5,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.version = int(version)
        if compression is None:
            compression = MtgJsonCompression.GZIP
        self.compression = MtgJsonCompression(compression)

    def set_compression(self, compression=MtgJsonCompression):
        self.compression = MtgJsonCompression(compression)

    async def all_printings_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards, including all printings and variations, categorized by
        set.

        :return: An async iterator of CardSet
        """
        async for k, item in self._iterate_model(
            kind="AllPrintings", model=Set, mode=MtgJsonMode.DICT_OF_MODEL
        ):
            yield item

    all_printings = sync_generator(all_printings_async)

    async def all_identifiers_async(self) -> AsyncGenerator[Card, None]:
        """
        all Card (Set) cards organized by card UUID.

        :return: An async iterator of Card object (either CardToken, or CardSet)
        """
        async for k, item in self._iterate_model(  # type: ignore # no-qa
            kind="AllIdentifiers", model=Card
        ):
            yield item

    all_identifiers = sync_generator(all_identifiers_async)

    async def all_prices_async(self) -> AsyncGenerator[CardPrices, None]:
        """
        all prices of cards in various formats.

        :return: An async iterator of CardPrices
        """
        async for k, item in self._iterate_model(kind="AllPrices"):
            yield CardPrices(uuid=k, **item)  # type: ignore

    all_prices = sync_generator(all_prices_async)

    async def atomic_cards_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        every Card (Atomic) card.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="AtomicCards"):
            yield item

    atomic_cards = sync_generator(atomic_cards_async)

    async def card_types_async(self) -> CardTypes:
        """
        every card type of any type of card.

        :return: A ``CardTypes`` object
        """
        return await self._get_item("CardTypes", model=CardTypes)

    card_types = synchronize(card_types_async)

    async def compiled_list_async(self) -> List[str]:
        """
        all individual outputs from MTGJSON, such as AllPrintings, CardTypes, etc.

        :return: A list of string
        """
        return await self._get_item("CompiledList", model=list)

    compiled_list = synchronize(compiled_list_async)

    async def deck_list_async(self) -> AsyncGenerator[DeckList, None]:
        """
        all individual Deck data.

        :return: An async iterator of DeckList
        """
        async for i, item in self._iterate_model(
            kind="DeckList", model=DeckList, mode=MtgJsonMode.LIST_OF_MODEL
        ):
            yield item

    deck_list = sync_generator(deck_list_async)

    async def deck_async(self, file_name: str) -> Deck:
        """
        Recovers a deck data

        :param file_name: the deck file_name
        :return: A ``Deck`` object
        """
        return await self._get_item(f"decks/{file_name}", model=Deck)

    deck = synchronize(deck_async)

    async def enum_values_async(self) -> dict:
        """
        All known property values for various Data Models.

        :return: a ``dict`` object
        """
        return await self._get_item("EnumValues", model=dict)

    enum_values = synchronize(enum_values_async)

    async def keywords_async(self) -> Keywords:
        """
        a list of possible all keywords used on all cards.

        :return: A ``Keywords`` object
        """
        return await self._get_item("Keywords", model=Keywords)

    keywords = synchronize(keywords_async)

    async def legacy_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Legacy format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Legacy", model=Set):
            yield item

    legacy = sync_generator(legacy_async)

    async def legacy_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Legacy format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="LegacyAtomic"):
            yield item

    legacy_atomic = sync_generator(legacy_atomic_async)

    async def meta_async(self) -> Meta:
        """
        the metadata object with ISO 8601 dates for latest build and SemVer
        specifications of the MTGJSON release.

        :return: A Meta object
        """
        return await self._get_item("Meta", model=Meta)

    meta = synchronize(meta_async)

    async def modern_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Modern format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Modern", model=Set):
            yield item

    modern = sync_generator(modern_async)

    async def modern_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Modern format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="ModernAtomic"):
            yield item

    modern_atomic = sync_generator(modern_atomic_async)

    async def pauper_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Pauper format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="PauperAtomic"):
            yield item

    pauper_atomic = sync_generator(pauper_atomic_async)

    async def pioneer_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to cards legal in the
        Pioneer format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Pioneer", model=Set):
            yield item

    pioneer = sync_generator(pioneer_async)

    async def pioneer_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Pioneer format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="PioneerAtomic"):
            yield item

    pioneer_atomic = sync_generator(pioneer_atomic_async)

    async def set_list_async(self) -> AsyncGenerator[SetList, None]:
        """
        a list of meta data for all Set data.

        :return: An async iterator of ``SetList``
        """
        async for k, item in self._iterate_model(
            kind="SetList", model=SetList, mode=MtgJsonMode.LIST_OF_MODEL
        ):
            yield item

    set_list = sync_generator(set_list_async)

    async def set_async(self, code: str) -> SetList:
        """
        Get a Set data

        :param code: The set identifier, such as "IKO" for "Ikoria, lair of the
                     monsters"

        :return: The set representation
        """
        return await self._get_item(code, SetList)

    set = synchronize(set_async)

    async def standard_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to cards legal in the
        Standard format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Standard", model=Set):
            yield item

    standard = sync_generator(standard_async)

    async def standard_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Atomic) cards, restricted to cards legal in the Standard format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="StandardAtomic"):
            yield item

    standard_atomic = sync_generator(standard_atomic_async)

    async def tcg_player_skus_async(self) -> AsyncGenerator[TcgPlayerSKUs, None]:
        """
        TCGplayer SKU information based on card UUIDs.

        :return: an async iterator of ``TcgPlayerSKUs``
        """
        group: Optional[TcgPlayerSKUs] = None
        async for (k, i), item in self._iterate_model(
            kind="TcgplayerSkus",
            model=TcgPlayerSKU,
            mode=MtgJsonMode.DICT_OF_LIST_OF_MODEL,
        ):
            if not group or k != group.id:
                if group:
                    yield group
                group = TcgPlayerSKUs(uuid=k, skus=[])  # type: ignore
            group.skus.append(item)

            yield group

    tcg_player_skus = sync_generator(tcg_player_skus_async)

    async def vintage_async(self) -> AsyncGenerator[Set, None]:
        """
        all Card (Set) cards organized by Set, restricted to sets legal in the
        Vintage format.

        :return: An async iterator of ``Set``
        """
        async for k, item in self._iterate_model(kind="Vintage", model=Set):
            yield item

    vintage = sync_generator(vintage_async)

    async def vintage_atomic_async(self) -> AsyncGenerator[CardAtomic, None]:
        """
        all Card (Atomic) cards, restricted to sets legal in the Vintage format.

        :return: An async iterator of ``CardAtomic``
        """
        async for item in self._atomic(kind="VintageAtomic"):
            yield item

    vintage_atomic = sync_generator(vintage_atomic_async)

    async def _atomic(self, kind: str) -> AsyncGenerator[CardAtomic, None]:
        card: Optional[CardAtomic] = None
        async for (k, i), item in self._iterate_model(
            kind=kind, model=CardFace, mode=MtgJsonMode.DICT_OF_LIST_OF_MODEL
        ):
            if not card or k != card.ascii_name:
                if card:
                    yield card
                card = CardAtomic(asciiName=k, faces=[])  # type: ignore
            card.faces.append(item)

        if card:
            yield card

    async def _get_item(
        self,
        kind: str,
        model: Union[Type[_T], Type[Dict], Type[List]] = dict,
        **kwargs,
    ) -> _T:
        path = f"/api/v{self.version}/{kind}.json"
        if self.compression != MtgJsonCompression.NONE:
            path += "." + self.compression.value

        try:
            async with self.client.stream("GET", path, **kwargs) as f:
                async with compressor.open(
                    f.aiter_bytes(),
                    compression=self.compression.to_stream_compression(),
                ) as f2:
                    data = json.loads(await f2.read())
                    data = data.get("data")

                    if issubclass(model, MightstoneModel):
                        return model.model_validate(data)  # type: ignore

                    return model(data)  # type: ignore
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {model} data, {e.errors()}",
                url=path,
                method="GET",
                status=None,
                data=e,
            )
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from MTG Json",
                url=e.request.url,
                method=e.request.method,
                status=e.response.status_code,
                data=None,
            )

    @overload
    def _iterate_model(
        self, kind: str
    ) -> AsyncGenerator[Tuple[DictOfKey, Dict], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: None
    ) -> AsyncGenerator[Tuple[DictOfKey, Dict], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: Type[_T]
    ) -> AsyncGenerator[Tuple[DictOfKey, _T], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: None, mode: Literal[MtgJsonMode.DICT_OF_MODEL]
    ) -> AsyncGenerator[Tuple[DictOfKey, Dict], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: Type[_T], mode: Literal[MtgJsonMode.DICT_OF_MODEL]
    ) -> AsyncGenerator[Tuple[DictOfKey, _T], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: None, mode: Literal[MtgJsonMode.LIST_OF_MODEL]
    ) -> AsyncGenerator[Tuple[ListOfKey, Dict], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: Type[_T], mode: Literal[MtgJsonMode.LIST_OF_MODEL]
    ) -> AsyncGenerator[Tuple[ListOfKey, _T], None]: ...

    @overload
    def _iterate_model(
        self, kind: str, model: None, mode: Literal[MtgJsonMode.DICT_OF_LIST_OF_MODEL]
    ) -> AsyncGenerator[Tuple[DictOfListOfKey, Dict], None]: ...

    @overload
    def _iterate_model(
        self,
        kind: str,
        model: Type[_T],
        mode: Literal[MtgJsonMode.DICT_OF_LIST_OF_MODEL],
    ) -> AsyncGenerator[Tuple[DictOfListOfKey, _T], None]: ...

    async def _iterate_model(
        self,
        kind: str,
        model: Optional[Type[_T]] = None,
        mode: MtgJsonMode = MtgJsonMode.DICT_OF_MODEL,
        error_threshold: int = 10,
        **kwargs,
    ) -> AsyncGenerator[Tuple[GeneratorKey, Union[_T, Dict]], None]:
        error = 0
        async for k, v in self._iterate_raw(kind, mode, **kwargs):
            try:
                if model:
                    yield k, model.model_validate(v)
                else:
                    yield k, dict(v)
            except ValidationError as e:
                error += 1
                logger.warning(
                    "Failed to validate %s data, for item %s, %s", model, k, e.errors
                )

                if error > error_threshold:
                    raise RuntimeError(
                        "Too many model validation error, something is wrong"
                    )

    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.DICT_OF_MODEL],
    # ) -> AsyncGenerator[DictOfModel, None]: ...
    #
    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.DICT_OF_MODEL],
    #     ijson_path: str
    # ) -> AsyncGenerator[DictOfModel, None]: ...
    #
    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.DICT_OF_LIST_OF_MODEL],
    # ) -> AsyncGenerator[DictOfListOfModel, None]: ...
    #
    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.DICT_OF_LIST_OF_MODEL],
    #     ijson_path: str
    # ) -> AsyncGenerator[DictOfListOfModel, None]: ...
    #
    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.LIST_OF_MODEL],
    # ) -> AsyncGenerator[ListOfModel, None]: ...
    #
    # @overload
    # def _iterate_raw(
    #     self,
    #     kind: str,
    #     mode: Literal[MtgJsonMode.LIST_OF_MODEL],
    #     ijson_path: str
    # ) -> AsyncGenerator[ListOfModel, None]: ...

    async def _iterate_raw(
        self,
        kind: str,
        mode: MtgJsonMode,
        ijson_path: Optional[str] = None,
    ) -> AsyncGenerator[GeneratorModel, None]:
        path = f"/api/v{self.version}/{kind}.json"
        if self.compression != MtgJsonCompression.NONE:
            path += "." + self.compression.value

        compression = self.compression.to_stream_compression()

        generator: Callable[
            [Any, Any, Optional[str]], AsyncGenerator[GeneratorModel, None]
        ] = dict_of_model_generator
        if mode == MtgJsonMode.LIST_OF_MODEL:
            generator = list_of_model_generator
        elif mode == MtgJsonMode.DICT_OF_LIST_OF_MODEL:
            generator = dict_of_list_of_model_generator

        async with self.client.stream("GET", path) as f:
            try:
                f.raise_for_status()
            except HTTPStatusError as e:
                raise ServiceError(
                    message="Failed to fetch data from Mtg JSON",
                    url=e.request.url,
                    method=e.request.method,
                    status=e.response.status_code,
                    data=e.response.content,
                )

            async with compressor.open(f.aiter_bytes(), compression=compression) as bit:
                async for item in generator(self.ijson, bit, ijson_path):
                    yield item


async def dict_of_list_of_model_generator(
    ijson, bytes_iterator, ijson_path=None
) -> AsyncGenerator[DictOfListOfModel, None]:
    if not ijson_path:
        ijson_path = "data"
    async for k, line in ijson.kvitems_async(bytes_iterator, ijson_path):
        for i, value in enumerate(line, start=1):
            yield (k, i), value


async def list_of_model_generator(
    ijson, bytes_iterator, ijson_path=None
) -> AsyncGenerator[ListOfModel, None]:
    if not ijson_path:
        ijson_path = "data.item"
    async for i, v in asyncstdlib.enumerate(
        ijson.items_async(bytes_iterator, ijson_path)
    ):
        yield i, v


async def dict_of_model_generator(
    ijson, bytes_iterator, ijson_path=None
) -> AsyncGenerator[DictOfModel, None]:
    if not ijson_path:
        ijson_path = "data"
    async for k, v in ijson.kvitems_async(bytes_iterator, ijson_path):
        yield k, v



import click

from mightstone.ass import aiterator_to_list
from mightstone.cli.models import MightstoneCli, pass_mightstone
from mightstone.cli.utils import catch_service_error, pretty_print
from mightstone.services.scryfall.models import (
    CardIdentifierPath,
    CatalogType,
    RulingIdentifierPath,
)


@click.group()
def scryfall():
    pass


@scryfall.command(name="sets")
@click.option("--limit", type=int)
@catch_service_error
@pass_mightstone
def scryfall_sets(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.scryfall.sets_async(**kwargs)),
        cli.format,
    )


@scryfall.command(name="set")
@click.argument("id_or_code", type=str)
@pass_mightstone
def scryfall_set(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.set(**kwargs), cli.format)


@scryfall.command()
@click.argument("id", type=str)
@click.argument("type", type=click.Choice([t.value for t in CardIdentifierPath]))
@catch_service_error
@pass_mightstone
def card(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.card(**kwargs), cli.format)


@scryfall.command()
@click.argument("q", type=str)
@click.option("--limit", type=int, default=100)
@catch_service_error
@pass_mightstone
def search(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.scryfall.search_async(**kwargs)),
        cli.format,
    )


@scryfall.command()
@click.argument("q", type=str)
@catch_service_error
@pass_mightstone
def random(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.random(**kwargs), cli.format)


@scryfall.command()
@click.argument("q", type=str)
@click.option("--exact", type=bool, is_flag=True)
@click.option("--set", type=str)
@catch_service_error
@pass_mightstone
def named(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.named(**kwargs), cli.format)


@scryfall.command()
@click.argument("q", type=str)
@click.option("--include_extras", type=bool, is_flag=True)
@catch_service_error
@pass_mightstone
def autocomplete(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.autocomplete(**kwargs)), cli.format


class ScryfallIdentifier(click.ParamType):
    name = "identifier"

    def convert(self, value, param, ctx):
        item = {}
        for constraint in value.split(","):
            (key, value) = constraint.split(":", 1)
            item[key] = value
        return item


@scryfall.command()
@click.argument("identifiers", nargs=-1, type=ScryfallIdentifier())
@catch_service_error
@pass_mightstone
def collection(cli: MightstoneCli, **kwargs):
    """
    scryfall collection id:683a5707-cddb-494d-9b41-51b4584ded69 "name:Ancient tomb"
    "set:dmu,collector_number:150"

    :param obj:
    :param kwargs:
    :return:
    """
    pretty_print(
        aiterator_to_list(cli.app.scryfall.collection_async(**kwargs)),
        cli.format,
    )


@scryfall.command()
@click.argument("id", type=str)
@click.argument("type", type=click.Choice([t.value for t in RulingIdentifierPath]))
@click.option("-l", "--limit", type=int)
@catch_service_error
@pass_mightstone
def rulings(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.scryfall.rulings_async(**kwargs)),
        cli.format,
    )


@scryfall.command()
@click.option("-l", "--limit", type=int, required=False)
@catch_service_error
@pass_mightstone
def symbols(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.scryfall.symbols_async(**kwargs)),
        cli.format,
    )


@scryfall.command()
@click.argument("cost", type=str)
@catch_service_error
@pass_mightstone
def parse_mana(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.parse_mana(**kwargs), cli.format)


@scryfall.command()
@click.argument("type", type=click.Choice([t.value for t in CatalogType]))
@catch_service_error
@pass_mightstone
def catalog(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.catalog(**kwargs), cli.format)


@scryfall.command()
@click.option("-l", "--limit", type=int, default=100)
@catch_service_error
@pass_mightstone
def migrations(cli: MightstoneCli, **kwargs):
    pretty_print(
        aiterator_to_list(cli.app.scryfall.migrations_async(**kwargs)),
        cli.format,
    )


@scryfall.command()
@click.argument("id", type=str)
@catch_service_error
@pass_mightstone
def migration(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.scryfall.migration(**kwargs), cli.format)



from enum import Enum

import click
from pydantic import BaseModel, ConfigDict, Field

from mightstone.app import Mightstone


class CliFormat(str, Enum):
    JSON = "json"
    YAML = "yaml"


class MightstoneCli(BaseModel):
    """
    Command line CLI context
    """

    format: CliFormat = CliFormat.JSON
    app: Mightstone = Field(default_factory=Mightstone)
    model_config = ConfigDict(arbitrary_types_allowed=True)


pass_mightstone = click.make_pass_decorator(MightstoneCli, ensure=True)



import click

from mightstone.cli.models import MightstoneCli, pass_mightstone
from mightstone.cli.utils import pretty_print


@click.group()
@click.option("--cache", type=int, default=0)
def cardconjurer():
    pass


@cardconjurer.command()
@pass_mightstone
@click.argument("url_or_path")
def card(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.card_conjurer.card(**kwargs), cli.format)


@cardconjurer.command()
@pass_mightstone
@click.argument("url_or_path")
def template(cli: MightstoneCli, **kwargs):
    pretty_print(cli.app.card_conjurer.template(**kwargs), cli.format)


@cardconjurer.command()
@pass_mightstone
@click.argument("url_or_path")
@click.argument("output", type=click.File("wb"))
@click.option("--asset-root-url", type=str)
def render(cli: MightstoneCli, url_or_path, output, asset_root_url):
    card = cli.app.card_conjurer.card(url_or_path)
    if asset_root_url:
        card.asset_root_url = asset_root_url
    cli.app.card_conjurer.render(card, output)



import mightstone.services.scryfall.models

from .api import Scryfall
from .models import (
    Card,
    Catalog,
    Error,
    ManaCost,
    Migration,
    Ruling,
    SerializableCard,
    SerializableCatalog,
    SerializableMigration,
    SerializableRuling,
    SerializableSet,
    SerializableSymbol,
    SerializableTag,
    Set,
    Symbol,
    Tag,
)



from typing import Any, Type, TypeVar, Union

from injector import Injector, Provider, Scope, ScopeDecorator, SingletonScope

"""
This file implements a variant of resource cleanup provided in this MR
https://github.com/python-injector/injector/pull/252/files
"""

T = TypeVar("T")


class CleanupScope(SingletonScope):
    def __init__(self, injector: Injector) -> None:
        super().__init__(injector)
        # We have singletons here, so never cache them twice, since otherwise
        # the cleanup method might be invoked twice.
        self.cachedProviders: set[Provider[Any]] = set()

    def get(self, key: Type[T], provider_: Provider[T]) -> Provider[T]:
        obj = super().get(key, provider_)
        self.cachedProviders.add(obj)
        return obj


class CleanupInjector(Injector):
    def cleanup(self):
        cleanup_scope = self.get(CleanupScope)
        for cached_provider in cleanup_scope.cachedProviders:
            obj = cached_provider.get(self)
            if hasattr(obj, "cleanup") and callable(obj.cleanup):
                obj.cleanup()


cleaned = ScopeDecorator(CleanupScope)
ScopeType = Union[ScopeDecorator, Type[Scope], None]



import logging
from typing import Optional, Union

import beanie
import httpx_cache
import mongomock_motor
import motor.motor_asyncio
from appdirs import AppDirs

from mightstone.config import DbImplem, MightstoneSettings
from mightstone.core import get_documents
from mightstone.injector import CleanupInjector
from mightstone.services.cardconjurer import CardConjurer
from mightstone.services.edhrec import EdhRecApi, EdhRecStatic
from mightstone.services.mtgjson import MtgJson
from mightstone.services.scryfall import Scryfall
from mightstone.services.wiki import Wiki
from mightstone.services.wotc import RuleExplorer
from mightstone.storage import Mongod

logger = logging.getLogger("mightstone")


class Mightstone:
    """
    A Mighstone instance

    Using python dependency injector, this class provides the services in an
    orderly fashion after setting up dependencies such as Beanie.
    """

    def __init__(self, config: Optional[Union[MightstoneSettings, dict]] = None):
        from .container import modules

        self.container = CleanupInjector(modules)
        if config:
            if isinstance(config, dict):
                config = MightstoneSettings.model_validate(config)
            self.container.binder.bind(MightstoneSettings, config)

    @classmethod
    async def with_beanie(
        cls, config: Optional[Union[MightstoneSettings, dict]] = None
    ):
        self = cls(config)
        await self.beanie_init()
        return self

    async def beanie_init(self) -> None:
        await beanie.init_beanie(
            database=self.mongo_database,
            document_models=get_documents(),
        )

    def __del__(self):
        self.container.cleanup()

    @property
    def config(self) -> MightstoneSettings:
        return self.container.get(MightstoneSettings)

    @property
    def app_dirs(self) -> AppDirs:
        return self.container.get(AppDirs)

    @property
    def mongo_client(
        self,
    ) -> Union[
        motor.motor_asyncio.AsyncIOMotorClient, mongomock_motor.AsyncMongoMockClient
    ]:
        if (
            self.container.get(MightstoneSettings).storage.implementation
            == DbImplem.FAKE
        ):
            return self.container.get(mongomock_motor.AsyncMongoMockClient)
        return self.container.get(motor.motor_asyncio.AsyncIOMotorClient)

    @property
    def mongo_database(
        self,
    ) -> Union[
        motor.motor_asyncio.AsyncIOMotorDatabase, mongomock_motor.AsyncMongoMockDatabase
    ]:
        if (
            self.container.get(MightstoneSettings).storage.implementation
            == DbImplem.FAKE
        ):
            return self.container.get(mongomock_motor.AsyncMongoMockDatabase)
        return self.container.get(motor.motor_asyncio.AsyncIOMotorDatabase)

    @property
    def mongo_server(self) -> Optional[Mongod]:
        return self.container.get(Mongod)

    @property
    def cache_transport(self) -> httpx_cache.AsyncCacheControlTransport:
        return self.container.get(httpx_cache.AsyncCacheControlTransport)

    @property
    def scryfall(self) -> Scryfall:
        return self.container.get(Scryfall)

    @property
    def mtg_json(self) -> MtgJson:
        return self.container.get(MtgJson)

    @property
    def card_conjurer(self) -> CardConjurer:
        return self.container.get(CardConjurer)

    @property
    def edhrec_api(self) -> EdhRecApi:
        return self.container.get(EdhRecApi)

    @property
    def edhrec_static(self) -> EdhRecStatic:
        return self.container.get(EdhRecStatic)

    @property
    def rule_explorer(self) -> RuleExplorer:
        return self.container.get(RuleExplorer)

    @property
    def wiki(self) -> Wiki:
        return self.container.get(Wiki)



from enum import Enum
from typing import Annotated, Optional

from pydantic_core._pydantic_core import Url

from ...core import Fallback, MightstoneModel
from .color import ColorAffinity, Identity


class AbilityType(str, Enum):
    AUTO = "auto"
    STATIC = "static"
    SPELL = "spell"
    TRIGGERED = "triggered"
    ACTIVATED = "activated"
    EVASION = "evasion"
    CHARACTERISTIC_DEFINING = "characteristic-defining"


class Ability(MightstoneModel):
    name: str
    types: list[AbilityType]
    rules: Optional[list[str]] = None
    wiki: Optional[Url] = None
    introduced: Optional[str] = None
    last_seen: Optional[str] = None
    has_cost: bool = False
    reminder: Optional[str] = None
    stats: dict[str, int] = {}
    storm: Annotated[Optional[int], Fallback(None)] = None

    @property
    def is_evergreen(self) -> bool:
        return self.last_seen == "Evergreen"

    def affinity(self) -> ColorAffinity:
        raise NotImplementedError


class AbilityList(MightstoneModel):
    abilities: list[Ability] = []



from pathlib import Path

from ...services.wotc import ComprehensiveRules

"""
The MTG ruleset
"""

rules = ComprehensiveRules.model_validate_json(
    Path(__file__).parent.joinpath("rules.json").read_bytes()
)



import logging
import os
import pathlib
from logging.handlers import RotatingFileHandler
from typing import Union

import click

from .. import Mightstone, __author__, __version__
from ..config import MightstoneSettings
from ..core import MightstoneError
from ..services.cardconjurer.commands import cardconjurer
from ..services.edhrec.commands import edhrec
from ..services.mtgjson.commands import mtgjson
from ..services.scryfall.commands import scryfall
from ..services.wiki.commands import wiki
from ..services.wotc.commands import wotc
from .models import CliFormat, MightstoneCli, pass_mightstone
from .utils import coro, pretty_print


@click.group()
@click.option(
    "-f",
    "--format",
    type=click.Choice([t.value for t in CliFormat]),
    default=CliFormat.JSON,
)
@click.option("-v", "--verbose", count=True)
@click.option("-l", "--log-level", default="ERROR")
@click.option(
    "config_file",
    "-c",
    "--config",
    type=click.Path(readable=True, exists=True),
    default=None,
)
@pass_mightstone
@coro
async def cli(
    mightstone: MightstoneCli,
    format,
    verbose,
    log_level,
    config_file: Union[str, bytes, os.PathLike[str]],
):
    mightstone.format = format

    if config_file:
        with open(config_file, encoding="utf-8") as fp:
            try:
                settings = MightstoneSettings.model_validate_json(fp.read())
                mightstone.app = Mightstone(config=settings)
            except MightstoneError as e:
                raise click.ClickException(str(e) + "\n" + str(e.__context__))

    await mightstone.app.beanie_init()

    if verbose:
        log_level = logging.WARNING
    if verbose > 1:
        log_level = logging.INFO
    if verbose > 2:
        log_level = logging.DEBUG

    log_directory = pathlib.Path(mightstone.app.app_dirs.user_log_dir)
    if not log_directory.exists():
        os.makedirs(log_directory)
    logging.basicConfig(
        handlers=[
            logging.StreamHandler(),
            RotatingFileHandler(
                log_directory.joinpath("mightstone.log"),
                maxBytes=100000,
                backupCount=10,
            ),
        ],
        level=log_level,
        format=(
            "[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s"
        ),
        datefmt="%Y-%m-%dT%H:%M:%S",
    )


@cli.command()
@pass_mightstone
def config(mightstone: MightstoneCli):
    """Dumps configuration"""

    pretty_print(mightstone.app.config)


@cli.command()
@click.option("-v", "--verbose", count=True)
def version(verbose):
    """Displays the version"""

    click.echo("Version: %s" % __version__)
    if verbose > 0:
        click.echo("Author: %s" % __author__)


cli.add_command(mtgjson)
cli.add_command(scryfall)
cli.add_command(edhrec)
cli.add_command(cardconjurer)
cli.add_command(wiki)
cli.add_command(wotc)
